# Copyright 2025 BMO Soluciones, S.A.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Views for managing loans and salary advances (Préstamos y Adelantos).

This module handles:
- Creating loan/advance requests
- Approving/rejecting loans
- Viewing payment schedules
- Exporting payment tables to Excel/PDF
- Tracking loan balances and payments
"""

from __future__ import annotations

from datetime import date
from decimal import Decimal
from io import BytesIO

from flask import (
    Blueprint,
    flash,
    redirect,
    render_template,
    request,
    url_for,
    send_file,
    Response,
)
from flask_login import current_user

from coati_payroll.model import (
    db,
    Adelanto,
    Empleado,
    Moneda,
    Deduccion,
    AdelantoAbono,
)
from coati_payroll.forms import (
    PrestamoForm,
    PrestamoApprovalForm,
    PagoExtraordinarioForm,
    CondonacionForm,
)
from coati_payroll.i18n import _
from coati_payroll.enums import AdelantoEstado, AdelantoTipo
from coati_payroll.rbac import require_read_access, require_write_access

prestamo_bp = Blueprint("prestamo", __name__, url_prefix="/prestamo")


@prestamo_bp.route("/")
@require_read_access()
def index():
    """List all loans and advances with filtering options."""
    # Get filter parameters
    empleado_id = request.args.get("empleado_id", "")
    estado = request.args.get("estado", "")
    tipo = request.args.get("tipo", "")

    # Build query
    query = db.select(Adelanto).join(Empleado)

    if empleado_id:
        query = query.filter(Adelanto.empleado_id == empleado_id)
    if estado:
        query = query.filter(Adelanto.estado == estado)
    if tipo:
        query = query.filter(Adelanto.tipo == tipo)

    # Order by most recent first
    query = query.order_by(Adelanto.fecha_solicitud.desc())

    prestamos = db.session.execute(query).scalars().all()

    # Get all employees for filter dropdown
    empleados = (
        db.session.execute(db.select(Empleado).filter_by(activo=True).order_by(Empleado.primer_apellido))
        .scalars()
        .all()
    )

    return render_template(
        "modules/prestamo/index.html",
        prestamos=prestamos,
        empleados=empleados,
        filtro_empleado=empleado_id,
        filtro_estado=estado,
        filtro_tipo=tipo,
        estados=AdelantoEstado,
        tipos=AdelantoTipo,
    )


@prestamo_bp.route("/new", methods=["GET", "POST"])
@require_write_access()
def new():
    """Create a new loan or salary advance."""
    form = PrestamoForm()

    # Populate select fields
    empleados = (
        db.session.execute(db.select(Empleado).filter_by(activo=True).order_by(Empleado.primer_apellido))
        .scalars()
        .all()
    )
    form.empleado_id.choices = [
        (emp.id, f"{emp.primer_nombre} {emp.primer_apellido} - {emp.codigo_empleado}") for emp in empleados
    ]

    monedas = db.session.execute(db.select(Moneda).filter_by(activo=True).order_by(Moneda.nombre)).scalars().all()
    form.moneda_id.choices = [(m.id, f"{m.nombre} ({m.codigo})") for m in monedas]

    deducciones = (
        db.session.execute(db.select(Deduccion).filter_by(activo=True).order_by(Deduccion.nombre)).scalars().all()
    )
    form.deduccion_id.choices = [("", _("-- Sin deducción asociada --"))] + [(d.id, d.nombre) for d in deducciones]

    if form.validate_on_submit():
        prestamo = Adelanto()
        prestamo.empleado_id = form.empleado_id.data
        prestamo.tipo = form.tipo.data
        prestamo.fecha_solicitud = form.fecha_solicitud.data
        prestamo.monto_solicitado = form.monto_solicitado.data
        prestamo.moneda_id = form.moneda_id.data
        prestamo.cuotas_pactadas = form.cuotas_pactadas.data
        prestamo.tasa_interes = form.tasa_interes.data or Decimal("0.0000")
        prestamo.tipo_interes = form.tipo_interes.data
        prestamo.metodo_amortizacion = form.metodo_amortizacion.data
        prestamo.cuenta_debe = form.cuenta_debe.data
        prestamo.cuenta_haber = form.cuenta_haber.data
        prestamo.motivo = form.motivo.data
        prestamo.estado = AdelantoEstado.BORRADOR
        prestamo.saldo_pendiente = Decimal("0.00")  # Will be set upon approval
        prestamo.creado_por = current_user.usuario

        # Set deduccion_id only if a valid value was selected
        if form.deduccion_id.data:
            prestamo.deduccion_id = form.deduccion_id.data

        db.session.add(prestamo)
        db.session.commit()

        flash(
            _("Préstamo/Adelanto creado exitosamente. Estado: Borrador."),
            "success",
        )
        return redirect(url_for("prestamo.detail", prestamo_id=prestamo.id))

    # Set default values
    if request.method == "GET":
        form.fecha_solicitud.data = date.today()
        form.tipo_interes.data = "ninguno"
        form.tasa_interes.data = Decimal("0.0000")
        form.metodo_amortizacion.data = "frances"

    return render_template("modules/prestamo/form.html", form=form, prestamo=None)


@prestamo_bp.route("/<prestamo_id>")
@require_read_access()
def detail(prestamo_id):
    """View loan details including payment schedule."""
    prestamo = db.session.get(Adelanto, prestamo_id)
    if not prestamo:
        flash("Préstamo no encontrado.", "danger")
        return redirect(url_for("prestamo.index"))

    # Touch relationship to ensure it is loaded before rendering
    prestamo.empleado

    # Generate payment schedule
    tabla_pago = generar_tabla_pago(prestamo)

    # Get payment history
    abonos = (
        db.session.execute(
            db.select(AdelantoAbono).filter_by(adelanto_id=prestamo_id).order_by(AdelantoAbono.fecha_abono.desc())
        )
        .scalars()
        .all()
    )

    # Get interest journal if loan has interest
    from coati_payroll.model import InteresAdelanto

    intereses = []
    if prestamo.tasa_interes and prestamo.tasa_interes > 0:
        intereses = (
            db.session.execute(
                db.select(InteresAdelanto)
                .filter_by(adelanto_id=prestamo_id)
                .order_by(InteresAdelanto.fecha_hasta.desc())
            )
            .scalars()
            .all()
        )

    return render_template(
        "modules/prestamo/detail.html",
        prestamo=prestamo,
        tabla_pago=tabla_pago,
        abonos=abonos,
        intereses=intereses,
    )


@prestamo_bp.route("/<prestamo_id>/edit", methods=["GET", "POST"])
@require_write_access()
def edit(prestamo_id):
    """Edit a loan (only allowed in draft or pending state)."""
    prestamo = db.session.get(Adelanto, prestamo_id)
    if not prestamo:
        flash(_("Préstamo no encontrado."), "danger")
        return redirect(url_for("prestamo.index"))

    # Only allow editing in draft or pending state
    if prestamo.estado not in [AdelantoEstado.BORRADOR, AdelantoEstado.PENDIENTE]:
        flash(
            _("No se puede editar un préstamo en estado '{estado}'.".format(estado=prestamo.estado)),
            "warning",
        )
        return redirect(url_for("prestamo.detail", prestamo_id=prestamo_id))

    form = PrestamoForm(obj=prestamo)

    # Populate select fields
    empleados = (
        db.session.execute(db.select(Empleado).filter_by(activo=True).order_by(Empleado.primer_apellido))
        .scalars()
        .all()
    )
    form.empleado_id.choices = [
        (emp.id, f"{emp.primer_nombre} {emp.primer_apellido} - {emp.codigo_empleado}") for emp in empleados
    ]

    monedas = db.session.execute(db.select(Moneda).filter_by(activo=True).order_by(Moneda.nombre)).scalars().all()
    form.moneda_id.choices = [(m.id, f"{m.nombre} ({m.codigo})") for m in monedas]

    deducciones = (
        db.session.execute(db.select(Deduccion).filter_by(activo=True).order_by(Deduccion.nombre)).scalars().all()
    )
    form.deduccion_id.choices = [("", _("-- Sin deducción asociada --"))] + [(d.id, d.nombre) for d in deducciones]

    if form.validate_on_submit():
        prestamo.empleado_id = form.empleado_id.data
        prestamo.tipo = form.tipo.data
        prestamo.fecha_solicitud = form.fecha_solicitud.data
        prestamo.monto_solicitado = form.monto_solicitado.data
        prestamo.moneda_id = form.moneda_id.data
        prestamo.cuotas_pactadas = form.cuotas_pactadas.data
        prestamo.tasa_interes = form.tasa_interes.data or Decimal("0.0000")
        prestamo.tipo_interes = form.tipo_interes.data
        prestamo.metodo_amortizacion = form.metodo_amortizacion.data
        prestamo.cuenta_debe = form.cuenta_debe.data
        prestamo.cuenta_haber = form.cuenta_haber.data
        prestamo.motivo = form.motivo.data
        prestamo.modificado_por = current_user.usuario

        # Set deduccion_id
        if form.deduccion_id.data:
            prestamo.deduccion_id = form.deduccion_id.data
        else:
            prestamo.deduccion_id = None

        db.session.commit()

        flash(_("Préstamo/Adelanto actualizado exitosamente."), "success")
        return redirect(url_for("prestamo.detail", prestamo_id=prestamo_id))

    return render_template("modules/prestamo/form.html", form=form, prestamo=prestamo)


@prestamo_bp.route("/<prestamo_id>/submit", methods=["POST"])
@require_write_access()
def submit(prestamo_id):
    """Submit a loan for approval (change from draft to pending)."""
    prestamo = db.session.get(Adelanto, prestamo_id)
    if not prestamo:
        flash(_("Préstamo no encontrado."), "danger")
        return redirect(url_for("prestamo.index"))

    if prestamo.estado != AdelantoEstado.BORRADOR:
        flash(_("Solo los préstamos en borrador pueden ser enviados."), "warning")
        return redirect(url_for("prestamo.detail", prestamo_id=prestamo_id))

    prestamo.estado = AdelantoEstado.PENDIENTE
    prestamo.modificado_por = current_user.usuario
    db.session.commit()

    flash(_("Préstamo enviado para aprobación."), "success")
    return redirect(url_for("prestamo.detail", prestamo_id=prestamo_id))


@prestamo_bp.route("/<prestamo_id>/approve", methods=["GET", "POST"])
@require_write_access()
def approve(prestamo_id):
    """Approve a loan and set it as active."""
    prestamo = db.session.get(Adelanto, prestamo_id)
    if not prestamo:
        flash(_("Préstamo no encontrado."), "danger")
        return redirect(url_for("prestamo.index"))

    if prestamo.estado not in [AdelantoEstado.PENDIENTE, AdelantoEstado.BORRADOR]:
        flash(_("Este préstamo no puede ser aprobado en su estado actual."), "warning")
        return redirect(url_for("prestamo.detail", prestamo_id=prestamo_id))

    form = PrestamoApprovalForm()

    if form.validate_on_submit():
        if form.aprobar.data:
            # Approve the loan
            prestamo.monto_aprobado = form.monto_aprobado.data
            prestamo.fecha_aprobacion = form.fecha_aprobacion.data
            prestamo.fecha_desembolso = form.fecha_desembolso.data
            prestamo.estado = AdelantoEstado.APROBADO
            prestamo.aprobado_por = current_user.usuario

            # Calculate installment amount based on amortization method
            if prestamo.cuotas_pactadas and prestamo.cuotas_pactadas > 0:
                from coati_payroll.interes_engine import calcular_cuota_frances

                tasa_interes = prestamo.tasa_interes or Decimal("0.0000")
                metodo = prestamo.metodo_amortizacion or "frances"

                # For French method, calculate constant payment
                # For German method, payment varies so we store the first payment
                if metodo == "frances":
                    prestamo.monto_por_cuota = calcular_cuota_frances(
                        prestamo.monto_aprobado, tasa_interes, prestamo.cuotas_pactadas
                    )
                else:
                    # For German method, store average payment for reference
                    # Actual payment will be calculated per installment
                    prestamo.monto_por_cuota = prestamo.monto_aprobado / prestamo.cuotas_pactadas
            else:
                prestamo.monto_por_cuota = prestamo.monto_aprobado

            # Set pending balance and initialize interest tracking
            prestamo.saldo_pendiente = prestamo.monto_aprobado
            prestamo.interes_acumulado = Decimal("0.00")
            prestamo.fecha_ultimo_calculo_interes = prestamo.fecha_aprobacion or date.today()
            prestamo.modificado_por = current_user.usuario

            db.session.commit()
            flash(_("Préstamo aprobado exitosamente."), "success")

        elif form.rechazar.data:
            # Reject the loan
            prestamo.estado = AdelantoEstado.RECHAZADO
            prestamo.rechazado_por = current_user.usuario
            prestamo.motivo_rechazo = form.motivo_rechazo.data
            prestamo.modificado_por = current_user.usuario

            db.session.commit()
            flash(_("Préstamo rechazado."), "info")

        return redirect(url_for("prestamo.detail", prestamo_id=prestamo_id))

    # Pre-populate form with loan data
    if request.method == "GET":
        form.monto_aprobado.data = prestamo.monto_solicitado
        form.fecha_aprobacion.data = date.today()

    return render_template("modules/prestamo/approve.html", form=form, prestamo=prestamo)


@prestamo_bp.route("/<prestamo_id>/cancel", methods=["POST"])
@require_write_access()
def cancel(prestamo_id):
    """Cancel a loan."""
    prestamo = db.session.get(Adelanto, prestamo_id)
    if not prestamo:
        flash(_("Préstamo no encontrado."), "danger")
        return redirect(url_for("prestamo.index"))

    if prestamo.estado in [AdelantoEstado.PAGADO, AdelantoEstado.CANCELADO]:
        flash(_("Este préstamo ya está finalizado."), "warning")
        return redirect(url_for("prestamo.detail", prestamo_id=prestamo_id))

    prestamo.estado = AdelantoEstado.CANCELADO
    prestamo.modificado_por = current_user.usuario
    db.session.commit()

    flash(_("Préstamo cancelado."), "info")
    return redirect(url_for("prestamo.detail", prestamo_id=prestamo_id))


@prestamo_bp.route("/<prestamo_id>/pago-extraordinario", methods=["GET", "POST"])
@require_write_access()
def pago_extraordinario(prestamo_id):
    """Register an extraordinary/manual payment on a loan."""
    prestamo = db.session.get(Adelanto, prestamo_id)
    if not prestamo:
        flash("Préstamo no encontrado.", "danger")
        return redirect(url_for("prestamo.index"))

    # Touch relationship to ensure it is loaded before rendering
    prestamo.empleado

    # Only allow payments on approved/active loans
    if prestamo.estado not in [AdelantoEstado.APROBADO, AdelantoEstado.APLICADO]:
        flash(
            _("Solo se pueden registrar pagos en préstamos aprobados o aplicados."),
            "warning",
        )
        return redirect(url_for("prestamo.detail", prestamo_id=prestamo_id))

    if prestamo.saldo_pendiente <= 0:
        flash(_("Este préstamo ya está totalmente pagado."), "info")
        return redirect(url_for("prestamo.detail", prestamo_id=prestamo_id))

    form = PagoExtraordinarioForm()

    if form.validate_on_submit():
        monto_abonado = form.monto_abonado.data

        # Validate payment amount
        if monto_abonado > prestamo.saldo_pendiente:
            flash(
                _("El monto del pago ({monto}) excede el saldo pendiente ({saldo}).").format(
                    monto=monto_abonado, saldo=prestamo.saldo_pendiente
                ),
                "warning",
            )
            return render_template(
                "modules/prestamo/pago_extraordinario.html",
                form=form,
                prestamo=prestamo,
            )

        # Record the payment
        abono = AdelantoAbono()
        abono.adelanto_id = prestamo.id
        abono.fecha_abono = form.fecha_abono.data
        abono.monto_abonado = monto_abonado
        abono.saldo_anterior = prestamo.saldo_pendiente
        abono.saldo_posterior = prestamo.saldo_pendiente - monto_abonado
        abono.tipo_abono = "manual"
        abono.observaciones = form.observaciones.data or "Pago extraordinario"
        # Audit trail information
        abono.tipo_comprobante = form.tipo_comprobante.data
        abono.numero_comprobante = form.numero_comprobante.data
        abono.referencia_bancaria = form.referencia_bancaria.data
        # Optional accounting entries
        abono.cuenta_debe = form.cuenta_debe.data
        abono.cuenta_haber = form.cuenta_haber.data
        abono.creado_por = current_user.usuario

        # Update loan balance
        prestamo.saldo_pendiente = abono.saldo_posterior
        prestamo.modificado_por = current_user.usuario

        # Apply payment according to selected method
        tipo_aplicacion = form.tipo_aplicacion.data

        # Calculate remaining installments (those not yet paid)
        total_abonado_previo = sum(a.monto_abonado for a in prestamo.abonos if a.tipo_abono in ["nomina", "manual"])
        cuotas_pagadas = 0
        if prestamo.monto_por_cuota and prestamo.monto_por_cuota > 0:
            cuotas_pagadas = int(total_abonado_previo / prestamo.monto_por_cuota)

        cuotas_pendientes = prestamo.cuotas_pactadas - cuotas_pagadas

        if tipo_aplicacion == "reducir_cuotas":
            # Option 1: Reduce number of installments, keep installment amount
            if prestamo.monto_por_cuota and prestamo.monto_por_cuota > 0:
                cuotas_a_eliminar = int(monto_abonado / prestamo.monto_por_cuota)
                # Store original values for the observation
                cuotas_originales = prestamo.cuotas_pactadas
                # Adjust total installments
                nueva_cuotas_pactadas = max(cuotas_pagadas, prestamo.cuotas_pactadas - cuotas_a_eliminar)
                prestamo.cuotas_pactadas = nueva_cuotas_pactadas

                abono.observaciones = (
                    f"{abono.observaciones or 'Pago extraordinario'} - "
                    f"Cuotas reducidas de {cuotas_originales} a {nueva_cuotas_pactadas}"
                )

        elif tipo_aplicacion == "reducir_monto":
            # Option 2: Reduce installment amount, keep number of installments
            if cuotas_pendientes > 0:
                # Recalculate installment amount based on remaining balance
                nueva_cuota = prestamo.saldo_pendiente / cuotas_pendientes
                monto_original = prestamo.monto_por_cuota
                prestamo.monto_por_cuota = nueva_cuota

                abono.observaciones = (
                    f"{abono.observaciones or 'Pago extraordinario'} - "
                    f"Monto por cuota reducido de {monto_original:.2f} a {nueva_cuota:.2f}"
                )

        # Check if loan is fully paid
        if prestamo.saldo_pendiente <= Decimal("0.01"):  # Allow for rounding
            prestamo.saldo_pendiente = Decimal("0.00")
            prestamo.estado = AdelantoEstado.PAGADO

        db.session.add(abono)
        db.session.commit()

        flash(
            _("Pago extraordinario registrado exitosamente."),
            "success",
        )
        return redirect(url_for("prestamo.detail", prestamo_id=prestamo_id))

    # Pre-populate form
    if request.method == "GET":
        form.fecha_abono.data = date.today()
        # Default to reducing installment amount (usually more beneficial for employee)
        form.tipo_aplicacion.data = "reducir_monto"

    return render_template(
        "modules/prestamo/pago_extraordinario.html",
        form=form,
        prestamo=prestamo,
    )


@prestamo_bp.route("/<prestamo_id>/condonacion", methods=["GET", "POST"])
@require_write_access()
def condonacion(prestamo_id):
    """Record a loan forgiveness/write-off (condonación de deuda)."""
    prestamo = db.session.get(Adelanto, prestamo_id)
    if not prestamo:
        flash(_("Préstamo no encontrado."), "danger")
        return redirect(url_for("prestamo.index"))

    # Only allow forgiveness on approved/active loans
    if prestamo.estado not in [AdelantoEstado.APROBADO, AdelantoEstado.APLICADO]:
        flash(
            _("Solo se pueden condonar préstamos aprobados o aplicados."),
            "warning",
        )
        return redirect(url_for("prestamo.detail", prestamo_id=prestamo_id))

    if prestamo.saldo_pendiente <= 0:
        flash(_("Este préstamo ya está totalmente pagado."), "info")
        return redirect(url_for("prestamo.detail", prestamo_id=prestamo_id))

    form = CondonacionForm()

    if form.validate_on_submit():
        monto_condonado = form.monto_condonado.data

        # Validate forgiveness amount
        if monto_condonado > prestamo.saldo_pendiente:
            flash(
                _("El monto a condonar ({monto}) excede el saldo pendiente ({saldo}).").format(
                    monto=monto_condonado, saldo=prestamo.saldo_pendiente
                ),
                "warning",
            )
            return render_template(
                "modules/prestamo/condonacion.html",
                form=form,
                prestamo=prestamo,
            )

        # Record the forgiveness as a special type of payment
        abono = AdelantoAbono()
        abono.adelanto_id = prestamo.id
        abono.fecha_abono = form.fecha_condonacion.data
        abono.monto_abonado = monto_condonado
        abono.saldo_anterior = prestamo.saldo_pendiente
        abono.saldo_posterior = prestamo.saldo_pendiente - monto_condonado
        abono.tipo_abono = "condonacion"

        # Store complete audit trail
        abono.autorizado_por = form.autorizado_por.data
        abono.documento_soporte = form.documento_soporte.data
        abono.referencia_documento = form.referencia_documento.data
        abono.justificacion = form.justificacion.data
        # Optional accounting entries
        abono.cuenta_debe = form.cuenta_debe.data
        abono.cuenta_haber = form.cuenta_haber.data

        # Build observation summary
        porcentaje = ""
        if form.porcentaje_condonado.data:
            porcentaje = f" ({form.porcentaje_condonado.data}%)"

        abono.observaciones = (
            f"CONDONACIÓN DE DEUDA{porcentaje} - "
            f"Autorizado por: {form.autorizado_por.data}. "
            f"Documento: {form.documento_soporte.data} - {form.referencia_documento.data}"
        )
        abono.creado_por = current_user.usuario

        # Update loan balance
        prestamo.saldo_pendiente = abono.saldo_posterior
        prestamo.modificado_por = current_user.usuario

        # If loan is fully forgiven/paid, mark as paid
        if prestamo.saldo_pendiente <= Decimal("0.01"):  # Allow for rounding
            prestamo.saldo_pendiente = Decimal("0.00")
            prestamo.estado = AdelantoEstado.PAGADO

        db.session.add(abono)
        db.session.commit()

        flash(
            _("Condonación de deuda registrada exitosamente. Monto condonado: {monto}").format(monto=monto_condonado),
            "success",
        )
        return redirect(url_for("prestamo.detail", prestamo_id=prestamo_id))

    # Pre-populate form
    if request.method == "GET":
        form.fecha_condonacion.data = date.today()
        form.monto_condonado.data = prestamo.saldo_pendiente

    return render_template(
        "modules/prestamo/condonacion.html",
        form=form,
        prestamo=prestamo,
    )


@prestamo_bp.route("/<prestamo_id>/tabla-pago/excel")
@require_read_access()
def export_excel(prestamo_id):
    """Export payment schedule to Excel."""
    prestamo = db.session.get(Adelanto, prestamo_id)
    if not prestamo:
        flash(_("Préstamo no encontrado."), "danger")
        return redirect(url_for("prestamo.index"))

    try:
        from openpyxl import Workbook
        from openpyxl.styles import Font, PatternFill
    except ImportError:
        flash(
            _("Excel export no disponible. Instale openpyxl."),
            "warning",
        )
        return redirect(url_for("prestamo.detail", prestamo_id=prestamo_id))

    # Generate payment schedule
    tabla_pago = generar_tabla_pago(prestamo)

    # Create workbook
    wb = Workbook()
    ws = wb.active
    ws.title = "Tabla de Pagos"

    # Header
    ws["A1"] = "TABLA DE PAGOS - PRÉSTAMO/ADELANTO"
    ws["A1"].font = Font(bold=True, size=14)
    ws.merge_cells("A1:E1")

    # Loan details
    row = 3
    ws[f"A{row}"] = "Empleado:"
    ws[f"B{row}"] = f"{prestamo.empleado.primer_nombre} {prestamo.empleado.primer_apellido}"
    row += 1
    ws[f"A{row}"] = "Tipo:"
    ws[f"B{row}"] = prestamo.tipo
    row += 1
    ws[f"A{row}"] = "Monto:"
    ws[f"B{row}"] = float(prestamo.monto_aprobado or prestamo.monto_solicitado)
    row += 1
    ws[f"A{row}"] = "Cuotas:"
    ws[f"B{row}"] = prestamo.cuotas_pactadas
    row += 2

    # Table header
    headers = ["#", "Fecha Estimada", "Cuota", "Interés", "Capital", "Saldo"]
    for col, header in enumerate(headers, start=1):
        cell = ws.cell(row=row, column=col, value=header)
        cell.font = Font(bold=True)
        cell.fill = PatternFill(start_color="366092", end_color="366092", fill_type="solid")
        cell.font = Font(bold=True, color="FFFFFF")

    # Table data
    for item in tabla_pago:
        row += 1
        ws.cell(row=row, column=1, value=item["numero"])
        ws.cell(row=row, column=2, value=item["fecha_estimada"])
        ws.cell(row=row, column=3, value=float(item["cuota"]))
        ws.cell(row=row, column=4, value=float(item["interes"]))
        ws.cell(row=row, column=5, value=float(item["capital"]))
        ws.cell(row=row, column=6, value=float(item["saldo"]))

    # Save to BytesIO
    output = BytesIO()
    wb.save(output)
    output.seek(0)

    # Use first 8 chars of ULID (26 chars total) for short filename
    filename = f"tabla_pago_{prestamo.empleado.codigo_empleado}_{prestamo.id[:8]}.xlsx"

    return send_file(
        output,
        mimetype="application/vnd.openxmlformats-officedocument.spreadsheetml.sheet",
        as_attachment=True,
        download_name=filename,
    )


@prestamo_bp.route("/<prestamo_id>/tabla-pago/pdf")
@require_read_access()
def export_pdf(prestamo_id):
    """Export payment schedule to PDF."""
    prestamo = db.session.get(Adelanto, prestamo_id)
    if not prestamo:
        flash(_("Préstamo no encontrado."), "danger")
        return redirect(url_for("prestamo.index"))

    # Generate payment schedule
    tabla_pago = generar_tabla_pago(prestamo)

    # Render HTML template for PDF
    html = render_template(
        "modules/prestamo/tabla_pago_pdf.html",
        prestamo=prestamo,
        tabla_pago=tabla_pago,
    )

    try:
        from flask_weasyprint import HTML, render_pdf

        pdf = render_pdf(HTML(string=html))
        # Use first 8 chars of ULID (26 chars total) for short filename
        filename = f"tabla_pago_{prestamo.empleado.codigo_empleado}_{prestamo.id[:8]}.pdf"

        return Response(
            pdf,
            mimetype="application/pdf",
            headers={"Content-Disposition": f"attachment; filename={filename}"},
        )
    except ImportError:
        flash(
            _("PDF export no disponible. Instale WeasyPrint."),
            "warning",
        )
        return redirect(url_for("prestamo.detail", prestamo_id=prestamo_id))


def generar_tabla_pago(prestamo: Adelanto) -> list[dict]:
    """Generate payment schedule for a loan.

    Args:
        prestamo: Loan object

    Returns:
        List of payment schedule items with fields:
        - numero: Payment number
        - fecha_estimada: Estimated payment date
        - cuota: Total payment amount
        - interes: Interest portion
        - capital: Principal portion
        - saldo: Remaining balance
    """
    if not prestamo.cuotas_pactadas or prestamo.cuotas_pactadas <= 0:
        return []

    monto_base = prestamo.monto_aprobado or prestamo.monto_solicitado
    if not monto_base or monto_base <= 0:
        return []

    # Determine start date
    fecha_inicio = prestamo.fecha_aprobacion or prestamo.fecha_solicitud or date.today()

    # Import interest engine
    from coati_payroll.interes_engine import generar_tabla_amortizacion

    # Get interest rate and type
    tasa_interes = prestamo.tasa_interes or Decimal("0.0000")
    tipo_interes = prestamo.tipo_interes or "ninguno"
    metodo_amortizacion = prestamo.metodo_amortizacion or "frances"

    # Generate amortization schedule using the interest engine
    cuotas = generar_tabla_amortizacion(
        principal=monto_base,
        tasa_anual=tasa_interes,
        num_cuotas=prestamo.cuotas_pactadas,
        fecha_inicio=fecha_inicio,
        metodo=metodo_amortizacion,
        tipo_interes=tipo_interes,
    )

    # Convert to dict format for template
    tabla = []
    for cuota in cuotas:
        tabla.append(
            {
                "numero": cuota.numero,
                "fecha_estimada": cuota.fecha_estimada,
                "cuota": cuota.cuota_total,
                "interes": cuota.interes,
                "capital": cuota.capital,
                "saldo": cuota.saldo,
            }
        )

    return tabla



# Copyright 2025 BMO Soluciones, S.A.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Base repository for data access."""

from __future__ import annotations

from abc import ABC, abstractmethod
from typing import Generic, TypeVar, Optional

from sqlalchemy.orm import Session

T = TypeVar("T")


class BaseRepository(ABC, Generic[T]):
    """Base repository for data access operations."""

    def __init__(self, session: Session):
        self.session = session

    @abstractmethod
    def get_by_id(self, id: str) -> Optional[T]:
        """Get entity by ID."""
        pass

    @abstractmethod
    def save(self, entity: T) -> T:
        """Save entity."""
        pass



# Copyright 2025 BMO Soluciones, S.A.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Employee CRUD routes."""

from __future__ import annotations

from datetime import date
from decimal import Decimal

from flask import Blueprint, flash, redirect, render_template, request, url_for
from flask_login import current_user

from coati_payroll.forms import EmployeeForm
from coati_payroll.i18n import _
from coati_payroll.model import CampoPersonalizado, Empleado, Moneda, db
from coati_payroll.rbac import require_read_access, require_write_access
from coati_payroll.vistas.constants import PER_PAGE

employee_bp = Blueprint("employee", __name__, url_prefix="/employee")


def get_currency_choices():
    """Get list of currencies for select fields."""
    currencies = db.session.execute(db.select(Moneda).filter_by(activo=True).order_by(Moneda.codigo)).scalars().all()
    return [("", _("Seleccionar..."))] + [(c.id, f"{c.codigo} - {c.nombre}") for c in currencies]


def get_empresa_choices():
    """Get list of companies for select fields."""
    from coati_payroll.model import Empresa

    empresas = (
        db.session.execute(db.select(Empresa).filter_by(activo=True).order_by(Empresa.razon_social)).scalars().all()
    )
    return [("", _("Seleccionar..."))] + [(e.id, f"{e.codigo} - {e.razon_social}") for e in empresas]


def get_custom_fields():
    """Get all active custom fields ordered by 'orden'."""
    return (
        db.session.execute(db.select(CampoPersonalizado).filter_by(activo=True).order_by(CampoPersonalizado.orden))
        .scalars()
        .all()
    )


def process_custom_fields_from_request(custom_fields):
    """Process custom field values from form request and return as dict.

    Args:
        custom_fields: List of CampoPersonalizado objects

    Returns:
        Dictionary with custom field names as keys and their converted values
    """
    datos_adicionales = {}
    for field in custom_fields:
        field_name = f"custom_{field.nombre_campo}"
        raw_value = request.form.get(field_name, "")

        match field.tipo_dato:
            case "texto":
                stripped = raw_value.strip() if raw_value else ""
                datos_adicionales[field.nombre_campo] = stripped or None
            case "entero":
                try:
                    datos_adicionales[field.nombre_campo] = int(raw_value) if raw_value else None
                except ValueError:
                    datos_adicionales[field.nombre_campo] = None
            case "decimal":
                try:
                    datos_adicionales[field.nombre_campo] = float(raw_value) if raw_value else None
                except ValueError:
                    datos_adicionales[field.nombre_campo] = None
            case "booleano":
                # Checkbox will send value only if checked
                datos_adicionales[field.nombre_campo] = field_name in request.form
            case _:
                # Unknown type, store as text
                datos_adicionales[field.nombre_campo] = raw_value or None
    return datos_adicionales


def process_last_three_salaries(form):
    """Process last three salary fields from form and return as dict.

    Stores salaries as strings to preserve Decimal precision in JSON.

    Args:
        form: EmployeeForm instance with salary fields

    Returns:
        Dictionary with last three salaries as strings, or None if empty
    """
    ultimos_salarios = {}
    if form.ultimo_salario_1.data:
        ultimos_salarios["salario_1"] = str(form.ultimo_salario_1.data)
    if form.ultimo_salario_2.data:
        ultimos_salarios["salario_2"] = str(form.ultimo_salario_2.data)
    if form.ultimo_salario_3.data:
        ultimos_salarios["salario_3"] = str(form.ultimo_salario_3.data)
    return ultimos_salarios if ultimos_salarios else None


@employee_bp.route("/")
@require_read_access()
def index():
    """List all employees with pagination."""
    page = request.args.get("page", 1, type=int)
    pagination = db.paginate(
        db.select(Empleado).order_by(Empleado.primer_apellido, Empleado.primer_nombre),
        page=page,
        per_page=PER_PAGE,
        error_out=False,
    )
    return render_template("modules/employee/index.html", employees=pagination.items, pagination=pagination)


@employee_bp.route("/new", methods=["GET", "POST"])
@require_write_access()
def new():
    """Create a new employee. Admin and HR can create employees."""
    form = EmployeeForm()
    form.moneda_id.choices = get_currency_choices()
    form.empresa_id.choices = get_empresa_choices()
    custom_fields = get_custom_fields()

    if form.validate_on_submit():
        employee = Empleado()
        # Set codigo_empleado only if provided (otherwise default will be used)
        if form.codigo_empleado.data and form.codigo_empleado.data.strip():
            employee.codigo_empleado = form.codigo_empleado.data.strip()
        employee.primer_nombre = form.primer_nombre.data
        employee.segundo_nombre = form.segundo_nombre.data
        employee.primer_apellido = form.primer_apellido.data
        employee.segundo_apellido = form.segundo_apellido.data
        employee.genero = form.genero.data or None
        employee.nacionalidad = form.nacionalidad.data
        employee.tipo_identificacion = form.tipo_identificacion.data or None
        employee.identificacion_personal = form.identificacion_personal.data
        employee.id_seguridad_social = form.id_seguridad_social.data or None
        employee.id_fiscal = form.id_fiscal.data or None
        employee.tipo_sangre = form.tipo_sangre.data or None
        employee.fecha_nacimiento = form.fecha_nacimiento.data
        employee.fecha_alta = form.fecha_alta.data
        employee.fecha_baja = form.fecha_baja.data
        employee.activo = form.activo.data
        employee.cargo = form.cargo.data
        employee.area = form.area.data
        employee.centro_costos = form.centro_costos.data
        employee.salario_base = form.salario_base.data or Decimal("0.00")
        employee.moneda_id = form.moneda_id.data or None
        employee.empresa_id = form.empresa_id.data or None
        employee.correo = form.correo.data
        employee.telefono = form.telefono.data
        employee.direccion = form.direccion.data
        employee.estado_civil = form.estado_civil.data or None
        employee.banco = form.banco.data
        employee.numero_cuenta_bancaria = form.numero_cuenta_bancaria.data
        employee.tipo_contrato = form.tipo_contrato.data or None
        employee.creado_por = current_user.usuario

        # Initial implementation data
        employee.anio_implementacion_inicial = form.anio_implementacion_inicial.data
        employee.mes_ultimo_cierre = form.mes_ultimo_cierre.data
        employee.salario_acumulado = form.salario_acumulado.data or Decimal("0.00")
        employee.impuesto_acumulado = form.impuesto_acumulado.data or Decimal("0.00")

        # Store last three salaries in JSON format using helper function
        employee.ultimos_tres_salarios = process_last_three_salaries(form)

        # Process custom fields
        employee.datos_adicionales = process_custom_fields_from_request(custom_fields)

        db.session.add(employee)
        db.session.commit()
        flash(_("Empleado creado exitosamente."), "success")
        return redirect(url_for("employee.index"))

    # Default date to today
    if not form.fecha_alta.data:
        form.fecha_alta.data = date.today()
    if not form.salario_base.data:
        form.salario_base.data = Decimal("0.00")

    return render_template(
        "modules/employee/form.html",
        form=form,
        title=_("Nuevo Empleado"),
        custom_fields=custom_fields,
        custom_values={},
    )


@employee_bp.route("/edit/<string:id>", methods=["GET", "POST"])
@require_write_access()
def edit(id: str):
    """Edit an existing employee. Admin and HR can edit employees."""
    employee = db.session.get(Empleado, id)
    if not employee:
        flash(_("Empleado no encontrado."), "error")
        return redirect(url_for("employee.index"))

    form = EmployeeForm(obj=employee)
    form.moneda_id.choices = get_currency_choices()
    form.empresa_id.choices = get_empresa_choices()
    custom_fields = get_custom_fields()

    if form.validate_on_submit():
        # Update codigo_empleado only if provided
        if form.codigo_empleado.data and form.codigo_empleado.data.strip():
            employee.codigo_empleado = form.codigo_empleado.data.strip()
        employee.primer_nombre = form.primer_nombre.data
        employee.segundo_nombre = form.segundo_nombre.data
        employee.primer_apellido = form.primer_apellido.data
        employee.segundo_apellido = form.segundo_apellido.data
        employee.genero = form.genero.data or None
        employee.nacionalidad = form.nacionalidad.data
        employee.tipo_identificacion = form.tipo_identificacion.data or None
        employee.identificacion_personal = form.identificacion_personal.data
        employee.id_seguridad_social = form.id_seguridad_social.data or None
        employee.id_fiscal = form.id_fiscal.data or None
        employee.tipo_sangre = form.tipo_sangre.data or None
        employee.fecha_nacimiento = form.fecha_nacimiento.data
        employee.fecha_alta = form.fecha_alta.data
        employee.fecha_baja = form.fecha_baja.data
        employee.activo = form.activo.data
        employee.cargo = form.cargo.data
        employee.area = form.area.data
        employee.centro_costos = form.centro_costos.data
        employee.salario_base = form.salario_base.data or Decimal("0.00")
        employee.moneda_id = form.moneda_id.data or None
        employee.empresa_id = form.empresa_id.data or None
        employee.correo = form.correo.data
        employee.telefono = form.telefono.data
        employee.direccion = form.direccion.data
        employee.estado_civil = form.estado_civil.data or None
        employee.banco = form.banco.data
        employee.numero_cuenta_bancaria = form.numero_cuenta_bancaria.data
        employee.tipo_contrato = form.tipo_contrato.data or None
        employee.modificado_por = current_user.usuario

        # Initial implementation data
        employee.anio_implementacion_inicial = form.anio_implementacion_inicial.data
        employee.mes_ultimo_cierre = form.mes_ultimo_cierre.data
        employee.salario_acumulado = form.salario_acumulado.data or Decimal("0.00")
        employee.impuesto_acumulado = form.impuesto_acumulado.data or Decimal("0.00")

        # Store last three salaries in JSON format using helper function
        employee.ultimos_tres_salarios = process_last_three_salaries(form)

        # Process custom fields
        employee.datos_adicionales = process_custom_fields_from_request(custom_fields)

        db.session.commit()
        flash(_("Empleado actualizado exitosamente."), "success")
        return redirect(url_for("employee.index"))

    # Pre-populate last three salaries from employee data
    if request.method != "POST":
        ultimos_salarios = employee.ultimos_tres_salarios or {}
        if ultimos_salarios.get("salario_1"):
            form.ultimo_salario_1.data = Decimal(str(ultimos_salarios["salario_1"]))
        if ultimos_salarios.get("salario_2"):
            form.ultimo_salario_2.data = Decimal(str(ultimos_salarios["salario_2"]))
        if ultimos_salarios.get("salario_3"):
            form.ultimo_salario_3.data = Decimal(str(ultimos_salarios["salario_3"]))

    # Get existing custom field values
    custom_values = employee.datos_adicionales or {}

    return render_template(
        "modules/employee/form.html",
        form=form,
        title=_("Editar Empleado"),
        employee=employee,
        custom_fields=custom_fields,
        custom_values=custom_values,
    )


@employee_bp.route("/delete/<string:id>", methods=["POST"])
@require_write_access()
def delete(id: str):
    """Delete an employee. Admin and HR can delete employees."""
    employee = db.session.get(Empleado, id)
    if not employee:
        flash(_("Empleado no encontrado."), "error")
        return redirect(url_for("employee.index"))

    db.session.delete(employee)
    db.session.commit()
    flash(_("Empleado eliminado exitosamente."), "success")
    return redirect(url_for("employee.index"))



# Copyright 2025 BMO Soluciones, S.A.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Settings page to consolidate administrative options."""

from __future__ import annotations

from flask import Blueprint, render_template

from coati_payroll.rbac import require_write_access

settings_bp = Blueprint("settings", __name__, url_prefix="/settings")


@settings_bp.route("/")
@require_write_access()
def index():
    """Display settings page with links to all configuration options."""
    return render_template("modules/settings/index.html")



# Copyright 2025 BMO Soluciones, S.A.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Loan processor for automatic loan and advance deductions."""

from __future__ import annotations

from datetime import date
from decimal import Decimal

from coati_payroll.model import db, Adelanto, AdelantoAbono, Nomina, Liquidacion
from coati_payroll.enums import AdelantoEstado
from coati_payroll.i18n import _
from ..domain.calculation_items import DeduccionItem


class LoanProcessor:
    """Processor for automatic loan and advance deductions."""

    def __init__(
        self,
        nomina: Nomina | None,
        fecha_calculo: date,
        periodo_inicio: date,
        periodo_fin: date,
        liquidacion: Liquidacion | None = None,
        calcular_interes: bool = True,
    ):
        self.nomina = nomina
        self.liquidacion = liquidacion
        self.fecha_calculo = fecha_calculo
        self.periodo_inicio = periodo_inicio
        self.periodo_fin = periodo_fin
        self.calcular_interes = calcular_interes

    def process_loans(
        self, empleado_id: str, saldo_disponible: Decimal, aplicar_prestamos: bool, prioridad_prestamos: int
    ) -> list[DeduccionItem]:
        """Process loans for an employee."""
        deductions = []

        if not aplicar_prestamos:
            return deductions

        # Get active loans
        from sqlalchemy import select

        prestamos = list(
            db.session.execute(
                select(Adelanto).filter(
                    Adelanto.empleado_id == empleado_id,
                    Adelanto.estado == AdelantoEstado.APROBADO,
                    Adelanto.saldo_pendiente > 0,
                    Adelanto.deduccion_id.isnot(None),  # Only loans, not advances
                )
            )
            .scalars()
            .all()
        )

        for prestamo in prestamos:
            if saldo_disponible <= 0:
                break

            # Calculate and apply interest if applicable
            if self.calcular_interes:
                self._calculate_interest(prestamo)

            monto_cuota = Decimal(str(prestamo.monto_por_cuota or 0))
            if monto_cuota <= 0:
                continue

            monto_aplicar = min(monto_cuota, saldo_disponible)

            item = DeduccionItem(
                codigo=f"PRESTAMO_{prestamo.id[:8]}",
                nombre=f"Cuota préstamo - {prestamo.motivo or 'N/A'}",
                monto=monto_aplicar,
                prioridad=prioridad_prestamos,
                es_obligatoria=False,
                tipo="prestamo",
            )
            deductions.append(item)
            saldo_disponible -= monto_aplicar

            # Record the payment
            self._record_payment(prestamo, monto_aplicar)

        return deductions

    def process_advances(
        self, empleado_id: str, saldo_disponible: Decimal, aplicar_adelantos: bool, prioridad_adelantos: int
    ) -> list[DeduccionItem]:
        """Process salary advances for an employee."""
        deductions = []

        if not aplicar_adelantos:
            return deductions

        from sqlalchemy import select

        adelantos = list(
            db.session.execute(
                select(Adelanto).filter(
                    Adelanto.empleado_id == empleado_id,
                    Adelanto.estado == AdelantoEstado.APROBADO,
                    Adelanto.saldo_pendiente > 0,
                    Adelanto.deduccion_id.is_(None),  # Only advances, not loans
                )
            )
            .scalars()
            .all()
        )

        for adelanto in adelantos:
            if saldo_disponible <= 0:
                break

            monto_cuota = Decimal(str(adelanto.monto_por_cuota or adelanto.saldo_pendiente))
            monto_aplicar = min(monto_cuota, saldo_disponible)

            item = DeduccionItem(
                codigo=f"ADELANTO_{adelanto.id[:8]}",
                nombre=f"Adelanto salarial - {adelanto.motivo or 'N/A'}",
                monto=monto_aplicar,
                prioridad=prioridad_adelantos,
                es_obligatoria=False,
                tipo="adelanto",
            )
            deductions.append(item)
            saldo_disponible -= monto_aplicar

            # Record the payment
            self._record_payment(adelanto, monto_aplicar)

        return deductions

    def _calculate_interest(self, prestamo: Adelanto) -> None:
        """Calculate and apply interest for a loan."""
        from coati_payroll.interes_engine import calcular_interes_periodo
        from coati_payroll.model import InteresAdelanto

        tasa_interes = prestamo.tasa_interes or Decimal("0.0000")
        if tasa_interes <= 0:
            return

        if prestamo.saldo_pendiente <= 0:
            return

        fecha_desde = prestamo.fecha_ultimo_calculo_interes
        if not fecha_desde:
            fecha_desde = prestamo.fecha_desembolso or prestamo.fecha_aprobacion

        if not fecha_desde:
            return

        fecha_hasta = self.fecha_calculo

        if fecha_desde >= fecha_hasta:
            return

        tipo_interes = prestamo.tipo_interes or "simple"
        interes_calculado, dias = calcular_interes_periodo(
            saldo=prestamo.saldo_pendiente,
            tasa_anual=tasa_interes,
            fecha_desde=fecha_desde,
            fecha_hasta=fecha_hasta,
            tipo_interes=tipo_interes,
        )

        if interes_calculado <= 0:
            return

        # Record interest in journal
        interes_entrada = InteresAdelanto(
            adelanto_id=prestamo.id,
            nomina_id=self.nomina.id if self.nomina else None,
            fecha_desde=fecha_desde,
            fecha_hasta=fecha_hasta,
            dias_transcurridos=dias,
            saldo_base=prestamo.saldo_pendiente,
            tasa_aplicada=tasa_interes,
            interes_calculado=interes_calculado,
            saldo_anterior=prestamo.saldo_pendiente,
            saldo_posterior=prestamo.saldo_pendiente + interes_calculado,
            observaciones=_("Interés calculado por nómina del {inicio} al {fin}").format(
                inicio=self.periodo_inicio, fin=self.periodo_fin
            ),
        )
        db.session.add(interes_entrada)

        # Update loan with interest
        prestamo.saldo_pendiente += interes_calculado
        prestamo.interes_acumulado = (prestamo.interes_acumulado or Decimal("0.00")) + interes_calculado
        prestamo.fecha_ultimo_calculo_interes = fecha_hasta

    def _record_payment(self, adelanto: Adelanto, monto: Decimal) -> None:
        """Record a payment towards a loan/advance."""
        saldo_anterior = Decimal(str(adelanto.saldo_pendiente))
        saldo_posterior = saldo_anterior - monto

        abono = AdelantoAbono(
            adelanto_id=adelanto.id,
            nomina_id=self.nomina.id if self.nomina else None,
            liquidacion_id=self.liquidacion.id if self.liquidacion else None,
            fecha_abono=self.fecha_calculo,
            monto_abonado=monto,
            saldo_anterior=saldo_anterior,
            saldo_posterior=max(saldo_posterior, Decimal("0.00")),
            tipo_abono="liquidacion" if self.liquidacion else "nomina",
        )
        db.session.add(abono)

        # Update adelanto balance
        adelanto.saldo_pendiente = max(saldo_posterior, Decimal("0.00"))
        if adelanto.saldo_pendiente <= 0:
            adelanto.estado = AdelantoEstado.PAGADO



# Copyright 2025 BMO Soluciones, S.A.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Concept calculator using Strategy pattern."""

from __future__ import annotations

from decimal import Decimal, ROUND_HALF_UP

from coati_payroll.enums import FormulaType
from coati_payroll.formula_engine import FormulaEngine, FormulaEngineError
from coati_payroll.model import db, Deduccion, ReglaCalculo
from ..domain.employee_calculation import EmpleadoCalculo


class ConceptCalculator:
    """Calculator for payroll concepts using Strategy pattern."""

    def __init__(self, config_repository, warnings: list[str]):
        self.config_repo = config_repository
        self.warnings = warnings

    def calculate(
        self,
        emp_calculo: EmpleadoCalculo,
        formula_tipo: str,
        monto_default: Decimal | None,
        porcentaje: Decimal | None,
        formula: dict | None,
        monto_override: Decimal | None,
        porcentaje_override: Decimal | None,
        codigo_concepto: str | None = None,
        base_calculo: str | None = None,
        unidad_calculo: str | None = None,
    ) -> Decimal:
        """Calculate concept amount."""
        # Use overrides if provided
        if monto_override:
            monto_calculado = Decimal(str(monto_override))
        elif porcentaje_override:
            monto_calculado = (emp_calculo.salario_base * Decimal(str(porcentaje_override)) / Decimal("100")).quantize(
                Decimal("0.01"), rounding=ROUND_HALF_UP
            )
        else:
            match formula_tipo:
                case FormulaType.FIJO:
                    monto_calculado = Decimal(str(monto_default or 0))

                case FormulaType.PORCENTAJE_SALARIO | FormulaType.PORCENTAJE:
                    if porcentaje:
                        monto_calculado = (
                            emp_calculo.salario_base * Decimal(str(porcentaje)) / Decimal("100")
                        ).quantize(Decimal("0.01"), rounding=ROUND_HALF_UP)
                    else:
                        monto_calculado = Decimal("0.00")

                case FormulaType.PORCENTAJE_BRUTO:
                    if porcentaje:
                        monto_calculado = (
                            emp_calculo.salario_bruto * Decimal(str(porcentaje)) / Decimal("100")
                        ).quantize(Decimal("0.01"), rounding=ROUND_HALF_UP)
                    else:
                        monto_calculado = Decimal("0.00")

                case FormulaType.HORAS:
                    monto_calculado = self._calculate_hours(emp_calculo, porcentaje, codigo_concepto, base_calculo)

                case FormulaType.DIAS:
                    monto_calculado = self._calculate_days(emp_calculo, porcentaje, codigo_concepto, base_calculo)

                case FormulaType.FORMULA:
                    monto_calculado = self._calculate_formula(emp_calculo, formula, codigo_concepto)

                case FormulaType.REGLA_CALCULO:
                    monto_calculado = self._calculate_regla_calculo(emp_calculo, codigo_concepto)

                case _:
                    monto_calculado = Decimal(str(monto_default or 0))

        # Ensure calculated amounts are never negative
        if monto_calculado < 0:
            self.warnings.append(
                f"Concepto '{codigo_concepto or 'desconocido'}': Configuración incorrecta resultó en "
                f"monto negativo ({monto_calculado}). Ajustando a 0.00. "
                f"Verifique la configuración del concepto (porcentaje o monto)."
            )
            return Decimal("0.00")

        return monto_calculado

    def _calculate_hours(
        self,
        emp_calculo: EmpleadoCalculo,
        porcentaje: Decimal | None,
        codigo_concepto: str | None,
        base_calculo: str | None,
    ) -> Decimal:
        """Calculate based on hours."""
        if not codigo_concepto or codigo_concepto not in emp_calculo.novedades:
            return Decimal("0.00")

        horas = emp_calculo.novedades[codigo_concepto]
        if horas <= 0:
            return Decimal("0.00")

        # Determine base for calculation
        if base_calculo == "salario_bruto":
            base = emp_calculo.salario_bruto
        else:
            base = emp_calculo.salario_mensual

        # Calculate hourly rate using configuration
        config = self.config_repo.get_for_empresa(emp_calculo.planilla.empresa_id)
        dias_base = Decimal(str(config.dias_mes_nomina))
        horas_dia = Decimal(str(config.horas_jornada_diaria))
        tasa_hora = (base / dias_base / horas_dia).quantize(Decimal("0.01"), rounding=ROUND_HALF_UP)

        # Apply percentage
        if porcentaje:
            tasa_hora = (tasa_hora * Decimal(str(porcentaje)) / Decimal("100")).quantize(
                Decimal("0.01"), rounding=ROUND_HALF_UP
            )

        # Calculate total for hours
        return (tasa_hora * horas).quantize(Decimal("0.01"), rounding=ROUND_HALF_UP)

    def _calculate_days(
        self,
        emp_calculo: EmpleadoCalculo,
        porcentaje: Decimal | None,
        codigo_concepto: str | None,
        base_calculo: str | None,
    ) -> Decimal:
        """Calculate based on days."""
        if not codigo_concepto or codigo_concepto not in emp_calculo.novedades:
            return Decimal("0.00")

        dias = emp_calculo.novedades[codigo_concepto]
        if dias <= 0:
            return Decimal("0.00")

        # Determine base for calculation
        if base_calculo == "salario_bruto":
            base = emp_calculo.salario_bruto
        else:
            base = emp_calculo.salario_mensual

        # Calculate daily rate using configuration
        config = self.config_repo.get_for_empresa(emp_calculo.planilla.empresa_id)
        dias_base = Decimal(str(config.dias_mes_nomina))
        tasa_dia = (base / dias_base).quantize(Decimal("0.01"), rounding=ROUND_HALF_UP)

        # Apply percentage
        if porcentaje:
            tasa_dia = (tasa_dia * Decimal(str(porcentaje)) / Decimal("100")).quantize(
                Decimal("0.01"), rounding=ROUND_HALF_UP
            )

        # Calculate total for days
        return (tasa_dia * dias).quantize(Decimal("0.01"), rounding=ROUND_HALF_UP)

    def _calculate_formula(
        self, emp_calculo: EmpleadoCalculo, formula: dict | None, codigo_concepto: str | None
    ) -> Decimal:
        """Calculate using formula engine."""
        if not formula or not isinstance(formula, dict):
            return Decimal("0.00")

        try:
            # Merge variables with formula inputs
            inputs = {**emp_calculo.variables_calculo}
            inputs["salario_bruto"] = emp_calculo.salario_bruto
            inputs["total_percepciones"] = emp_calculo.total_percepciones
            inputs["total_deducciones"] = emp_calculo.total_deducciones

            # Calculate before-tax deductions already processed in this period
            deducciones_antes_impuesto_periodo = Decimal("0.00")
            for ded in emp_calculo.deducciones:
                if ded.deduccion_id:
                    ded_obj = db.session.get(Deduccion, ded.deduccion_id)
                    if ded_obj and ded_obj.antes_impuesto:
                        deducciones_antes_impuesto_periodo += ded.monto
            inputs["deducciones_antes_impuesto_periodo"] = deducciones_antes_impuesto_periodo
            # Legacy alias for backward compatibility (deprecated but kept to avoid breaking existing schemas)
            inputs["inss_periodo"] = deducciones_antes_impuesto_periodo
            # New generic aliases (preferred for new schemas)
            inputs["pre_tax_deductions"] = deducciones_antes_impuesto_periodo
            inputs["social_security_deduction"] = deducciones_antes_impuesto_periodo

            engine = FormulaEngine(formula)
            result = engine.execute(inputs)
            return Decimal(str(result.get("output", 0))).quantize(Decimal("0.01"), rounding=ROUND_HALF_UP)
        except FormulaEngineError as e:
            self.warnings.append(f"Error en fórmula: {str(e)}")
            return Decimal("0.00")

    def _calculate_regla_calculo(self, emp_calculo: EmpleadoCalculo, codigo_concepto: str | None) -> Decimal:
        """Calculate using ReglaCalculo."""
        from sqlalchemy import select

        # Find the ReglaCalculo linked to this deduction
        regla = db.session.execute(
            select(ReglaCalculo).filter_by(deduccion_id=codigo_concepto).filter(ReglaCalculo.activo.is_(True))
        ).scalar_one_or_none()

        if not regla:
            # Try finding by deduccion_id matching deduccion's id
            deduccion_obj = db.session.execute(select(Deduccion).filter_by(codigo=codigo_concepto)).scalar_one_or_none()
            if deduccion_obj:
                regla = db.session.execute(
                    select(ReglaCalculo).filter_by(deduccion_id=deduccion_obj.id).filter(ReglaCalculo.activo.is_(True))
                ).scalar_one_or_none()

        if not regla or not regla.esquema_json:
            self.warnings.append(f"ReglaCalculo no encontrada para deducción {codigo_concepto}")
            return Decimal("0.00")

        try:
            # Prepare inputs for formula engine
            inputs = {**emp_calculo.variables_calculo}
            inputs["salario_bruto"] = emp_calculo.salario_bruto
            inputs["total_percepciones"] = emp_calculo.total_percepciones
            inputs["total_deducciones"] = emp_calculo.total_deducciones

            # Calculate before-tax deductions already processed
            deducciones_antes_impuesto_periodo = Decimal("0.00")
            for ded in emp_calculo.deducciones:
                if ded.deduccion_id:
                    ded_obj = db.session.get(Deduccion, ded.deduccion_id)
                    if ded_obj and ded_obj.antes_impuesto:
                        deducciones_antes_impuesto_periodo += ded.monto
            inputs["deducciones_antes_impuesto_periodo"] = deducciones_antes_impuesto_periodo
            # Legacy alias for backward compatibility (deprecated but kept to avoid breaking existing schemas)
            inputs["inss_periodo"] = deducciones_antes_impuesto_periodo
            # New generic aliases (preferred for new schemas)
            inputs["pre_tax_deductions"] = deducciones_antes_impuesto_periodo
            inputs["social_security_deduction"] = deducciones_antes_impuesto_periodo

            engine = FormulaEngine(regla.esquema_json)
            result = engine.execute(inputs)
            return Decimal(str(result.get("output", 0))).quantize(Decimal("0.01"), rounding=ROUND_HALF_UP)
        except FormulaEngineError as e:
            self.warnings.append(f"Error en ReglaCalculo {regla.codigo}: {str(e)}")
            return Decimal("0.00")



# Copyright 2025 BMO Soluciones, S.A.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Repository layer for data access."""

from .base_repository import BaseRepository
from .planilla_repository import PlanillaRepository
from .employee_repository import EmployeeRepository
from .acumulado_repository import AcumuladoRepository
from .novelty_repository import NoveltyRepository
from .exchange_rate_repository import ExchangeRateRepository
from .config_repository import ConfigRepository

__all__ = [
    "BaseRepository",
    "PlanillaRepository",
    "EmployeeRepository",
    "AcumuladoRepository",
    "NoveltyRepository",
    "ExchangeRateRepository",
    "ConfigRepository",
]



# Copyright 2025 BMO Soluciones, S.A.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Validator for Employee."""

from __future__ import annotations

from datetime import date
from decimal import Decimal

from coati_payroll.model import Empleado
from ..domain.payroll_context import PayrollContext
from ..results.validation_result import ValidationResult
from ..validators.base_validator import BaseValidator


class EmployeeValidator(BaseValidator):
    """Validates that an employee is eligible for payroll processing."""

    def validate(self, context: PayrollContext) -> ValidationResult:
        """Validate employee - this method signature is required by BaseValidator."""
        # This validator is called per-employee, not per-context
        # So we provide a separate method
        result = ValidationResult()
        return result

    def validate_employee(
        self, empleado: Empleado, planilla_empresa_id: str | None, periodo_inicio: date, periodo_fin: date
    ) -> ValidationResult:
        """Validate employee for payroll processing."""
        result = ValidationResult()

        if not empleado.activo:
            result.add_error(f"Empleado {empleado.codigo_empleado} no está activo")

        if empleado.fecha_alta:
            if empleado.fecha_alta > date.today():
                result.add_error(
                    f"Empleado {empleado.codigo_empleado}: fecha de ingreso ({empleado.fecha_alta}) "
                    f"es posterior a la fecha actual"
                )
            if empleado.fecha_alta > periodo_fin:
                result.add_error(
                    f"Empleado {empleado.codigo_empleado}: fecha de ingreso ({empleado.fecha_alta}) "
                    f"es posterior al período a procesar ({periodo_fin})"
                )
        else:
            result.add_error(f"Empleado {empleado.codigo_empleado} no tiene fecha de ingreso definida")

        if empleado.fecha_baja and empleado.fecha_baja < periodo_inicio:
            result.add_error(
                f"Empleado {empleado.codigo_empleado}: fecha de salida ({empleado.fecha_baja}) "
                f"es anterior al inicio del período ({periodo_inicio})"
            )

        if not empleado.identificacion_personal:
            result.add_error(f"Empleado {empleado.codigo_empleado} no tiene identificación personal")

        if empleado.salario_base <= Decimal("0.00"):
            result.add_error(
                f"Empleado {empleado.codigo_empleado} tiene salario base inválido ({empleado.salario_base})"
            )

        if not empleado.empresa_id:
            result.add_error(f"Empleado {empleado.codigo_empleado} no está asignado a ninguna empresa")

        if planilla_empresa_id and empleado.empresa_id:
            if empleado.empresa_id != planilla_empresa_id:
                result.add_error(
                    f"Empleado {empleado.codigo_empleado} pertenece a empresa diferente a la planilla. "
                    f"Empleado empresa_id={empleado.empresa_id}, Planilla empresa_id={planilla_empresa_id}"
                )

        if not empleado.moneda_id:
            result.add_error(f"Empleado {empleado.codigo_empleado} no tiene moneda definida")

        return result



# Copyright 2025 BMO Soluciones, S.A.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Initial data for currencies, income concepts, deduction concepts, and payroll types.

This module provides default data to be loaded during system initialization.
All data is jurisdiction-agnostic and uses Flask-Babel for translation support.
Strings are marked for translation using _() and will be translated based on
the configured language in the database.
"""

from __future__ import annotations

# <-------------------------------------------------------------------------> #
# Standard library
# <-------------------------------------------------------------------------> #

# <-------------------------------------------------------------------------> #
# Third party libraries
# <-------------------------------------------------------------------------> #

# <-------------------------------------------------------------------------> #
# Local modules
# <-------------------------------------------------------------------------> #
from coati_payroll.i18n import _l as _


# American currencies (North, Central, South America, and Caribbean)
# Currency names are marked for translation
CURRENCIES = [
    # North America
    {"codigo": "USD", "nombre": _("US Dollar"), "simbolo": "$"},
    {"codigo": "CAD", "nombre": _("Canadian Dollar"), "simbolo": "C$"},
    {"codigo": "MXN", "nombre": _("Mexican Peso"), "simbolo": "$"},
    # Central America
    {"codigo": "BZD", "nombre": _("Belize Dollar"), "simbolo": "BZ$"},
    {"codigo": "CRC", "nombre": _("Costa Rican Colón"), "simbolo": "₡"},
    {"codigo": "GTQ", "nombre": _("Guatemalan Quetzal"), "simbolo": "Q"},
    {"codigo": "HNL", "nombre": _("Honduran Lempira"), "simbolo": "L"},
    {"codigo": "NIO", "nombre": _("Nicaraguan Córdoba"), "simbolo": "C$"},
    {"codigo": "PAB", "nombre": _("Panamanian Balboa"), "simbolo": "B/."},
    {"codigo": "SVC", "nombre": _("Salvadoran Colón"), "simbolo": "₡"},
    # South America
    {"codigo": "ARS", "nombre": _("Argentine Peso"), "simbolo": "$"},
    {"codigo": "BOB", "nombre": _("Bolivian Boliviano"), "simbolo": "Bs."},
    {"codigo": "BRL", "nombre": _("Brazilian Real"), "simbolo": "R$"},
    {"codigo": "CLP", "nombre": _("Chilean Peso"), "simbolo": "$"},
    {"codigo": "COP", "nombre": _("Colombian Peso"), "simbolo": "$"},
    {"codigo": "GYD", "nombre": _("Guyanese Dollar"), "simbolo": "G$"},
    {"codigo": "PEN", "nombre": _("Peruvian Sol"), "simbolo": "S/"},
    {"codigo": "PYG", "nombre": _("Paraguayan Guaraní"), "simbolo": "₲"},
    {"codigo": "SRD", "nombre": _("Surinamese Dollar"), "simbolo": "$"},
    {"codigo": "UYU", "nombre": _("Uruguayan Peso"), "simbolo": "$"},
    {"codigo": "VES", "nombre": _("Venezuelan Bolívar"), "simbolo": "Bs."},
    # Caribbean
    {"codigo": "ANG", "nombre": _("Netherlands Antillean Guilder"), "simbolo": "ƒ"},
    {"codigo": "AWG", "nombre": _("Aruban Florin"), "simbolo": "ƒ"},
    {"codigo": "BBD", "nombre": _("Barbadian Dollar"), "simbolo": "$"},
    {"codigo": "BSD", "nombre": _("Bahamian Dollar"), "simbolo": "$"},
    {"codigo": "CUP", "nombre": _("Cuban Peso"), "simbolo": "$"},
    {"codigo": "DOP", "nombre": _("Dominican Peso"), "simbolo": "$"},
    {"codigo": "HTG", "nombre": _("Haitian Gourde"), "simbolo": "G"},
    {"codigo": "JMD", "nombre": _("Jamaican Dollar"), "simbolo": "J$"},
    {"codigo": "TTD", "nombre": _("Trinidad and Tobago Dollar"), "simbolo": "TT$"},
    {"codigo": "XCD", "nombre": _("East Caribbean Dollar"), "simbolo": "EC$"},
]


# Liquidation concepts (LiquidacionConcepto)
LIQUIDATION_CONCEPTS = [
    {
        "codigo": "DESPIDO",
        "nombre": _("Dismissal"),
        "descripcion": _("Termination initiated by employer"),
    },
    {
        "codigo": "RENUNCIA",
        "nombre": _("Resignation"),
        "descripcion": _("Termination initiated by employee"),
    },
]


# Income concepts (Percepciones / Ingresos)
# These add to employee salary
# All names and descriptions are marked for translation
INCOME_CONCEPTS = [
    {
        "codigo": "OVERTIME",
        "nombre": _("Overtime"),
        "descripcion": _("Additional hours worked beyond regular schedule"),
    },
    {
        "codigo": "COMMISSIONS",
        "nombre": _("Commissions"),
        "descripcion": _("Sales commissions and performance-based income"),
    },
    {
        "codigo": "BONUSES",
        "nombre": _("Bonuses"),
        "descripcion": _("Performance or achievement bonuses"),
    },
    {
        "codigo": "GOAL_INCENTIVES",
        "nombre": _("Goal Incentives"),
        "descripcion": _("Incentives for meeting specific targets or goals"),
    },
    {
        "codigo": "REPORTED_TIPS",
        "nombre": _("Reported Tips"),
        "descripcion": _("Tips and gratuities reported by employee"),
    },
    {
        "codigo": "TRANSPORT_ALLOWANCE",
        "nombre": _("Transportation Allowance"),
        "descripcion": _("Fixed transportation allowance"),
    },
    {
        "codigo": "FOOD_ALLOWANCE",
        "nombre": _("Food Allowance"),
        "descripcion": _("Fixed food or meal allowance"),
    },
    {
        "codigo": "HOUSING_ALLOWANCE",
        "nombre": _("Housing Allowance"),
        "descripcion": _("Fixed housing or rent allowance"),
    },
    {
        "codigo": "SENIORITY_BONUS",
        "nombre": _("Seniority Bonus"),
        "descripcion": _("Additional payment based on years of service"),
    },
    {
        "codigo": "PAID_VACATION",
        "nombre": _("Paid Vacation"),
        "descripcion": _("Payment for vacation time taken"),
    },
    {
        "codigo": "HOLIDAY_WORK",
        "nombre": _("Holiday Work"),
        "descripcion": _("Additional payment for working on holidays"),
    },
    {
        "codigo": "PAID_LEAVE",
        "nombre": _("Paid Leave"),
        "descripcion": _("Payment for authorized paid leave"),
    },
    {
        "codigo": "PAID_LICENSE",
        "nombre": _("Paid License"),
        "descripcion": _("Payment for extended paid license periods"),
    },
    {
        "codigo": "RETROACTIVE_PAYMENTS",
        "nombre": _("Retroactive Payments"),
        "descripcion": _("Back payments for salary adjustments"),
    },
    {
        "codigo": "THIRTEENTH_SALARY",
        "nombre": _("13th Month Salary"),
        "descripcion": _("Annual bonus (13th month salary)"),
    },
    {
        "codigo": "GRATUITIES",
        "nombre": _("Gratuities"),
        "descripcion": _("Discretionary bonuses or gratuities"),
    },
    {
        "codigo": "SEVERANCE_PAY",
        "nombre": _("Severance Pay"),
        "descripcion": _("Severance or termination compensation"),
    },
]


# Deduction concepts (Deducciones)
# These subtract from employee salary
# All names and descriptions are marked for translation
DEDUCTION_CONCEPTS = [
    {
        "codigo": "SALARY_ADVANCE",
        "nombre": _("Salary Advance"),
        "descripcion": _("Deduction for salary advance repayment (automatic from Loans/Advances module)"),
    },
    {
        "codigo": "VOLUNTARY_RETIREMENT",
        "nombre": _("Voluntary Retirement Plan"),
        "descripcion": _("Voluntary retirement savings contribution"),
    },
    {
        "codigo": "MEMBERSHIPS_SUBSCRIPTIONS",
        "nombre": _("Memberships/Subscriptions"),
        "descripcion": _("Deduction for memberships or subscription services"),
    },
    {
        "codigo": "CAFETERIA",
        "nombre": _("Cafeteria"),
        "descripcion": _("Company cafeteria charges"),
    },
    {
        "codigo": "INTERNAL_LOANS",
        "nombre": _("Internal Loans"),
        "descripcion": _("Internal loan repayment installments"),
    },
    {
        "codigo": "CREDIT_UNION_LOANS",
        "nombre": _("Credit Union Loans"),
        "descripcion": _("Credit union or cooperative loan payments"),
    },
    {
        "codigo": "LOAN_INTEREST",
        "nombre": _("Loan Interest"),
        "descripcion": _("Interest charges on loans"),
    },
    {
        "codigo": "ALIMONY",
        "nombre": _("Alimony"),
        "descripcion": _("Court-ordered alimony or child support payments"),
    },
    {
        "codigo": "COURT_GARNISHMENTS",
        "nombre": _("Court Garnishments"),
        "descripcion": _("Court-ordered wage garnishments"),
    },
    {
        "codigo": "UNPAID_ABSENCES",
        "nombre": _("Unpaid Absences"),
        "descripcion": _("Deductions for unpaid absences"),
    },
    {
        "codigo": "TARDINESS",
        "nombre": _("Tardiness"),
        "descripcion": _("Deductions for late arrivals"),
    },
    {
        "codigo": "NONCOMPLIANCE_PENALTIES",
        "nombre": _("Non-compliance Penalties"),
        "descripcion": _("Penalties for policy or contract violations"),
    },
    {
        "codigo": "LOSSES_DAMAGES",
        "nombre": _("Losses or Damages"),
        "descripcion": _("Deductions for lost or damaged property"),
    },
    {
        "codigo": "UNIFORM_PENALTIES",
        "nombre": _("Uniform Penalties"),
        "descripcion": _("Penalties for unreturned uniforms or equipment"),
    },
    {
        "codigo": "STORE_CANTEEN_PURCHASES",
        "nombre": _("Store/Canteen Purchases"),
        "descripcion": _("Purchases at company store or canteen"),
    },
    {
        "codigo": "VOLUNTARY_DONATIONS",
        "nombre": _("Voluntary Donations"),
        "descripcion": _("Voluntary charitable donations"),
    },
    {
        "codigo": "ASSOCIATION_CONTRIBUTIONS",
        "nombre": _("Association Contributions"),
        "descripcion": _("Voluntary contributions to associations"),
    },
    {
        "codigo": "UNION_DUES",
        "nombre": _("Union Dues"),
        "descripcion": _("Voluntary union membership dues"),
    },
    {
        "codigo": "OVERPAYMENT_CORRECTION",
        "nombre": _("Overpayment Correction"),
        "descripcion": _("Correction for previous overpayments"),
    },
]


# Employer benefit concepts (Prestaciones / Aportes Patronales)
# These are employer costs and provisions that do NOT affect employee's net pay
# All names and descriptions are marked for translation
BENEFIT_CONCEPTS = [
    {
        "codigo": "PAID_VACATION_PROVISION",
        "nombre": _("Paid Vacation Provision"),
        "descripcion": _("Employer provision for paid vacation days " "(universal labor right in the Americas)"),
    },
    {
        "codigo": "THIRTEENTH_SALARY_PROVISION",
        "nombre": _("13th Month Salary Provision"),
        "descripcion": _(
            "Employer provision for 13th month salary/Christmas bonus "
            "(aguinaldo - widespread right in Latin America)"
        ),
    },
    {
        "codigo": "SEVERANCE_PROVISION",
        "nombre": _("Severance Pay Provision"),
        "descripcion": _(
            "Employer provision for severance pay in case of unjust dismissal "
            "(common legal principle in the Americas)"
        ),
    },
]


# Payroll Types (Tipos de Planilla)
# Common payroll types with different periodicities
# All names and descriptions are marked for translation
PAYROLL_TYPES = [
    {
        "codigo": "MONTHLY",
        "descripcion": _("Monthly Payroll - 30 days"),
        "dias": 30,
        "periodicidad": "mensual",
        "periodos_por_anio": 12,
    },
    {
        "codigo": "BIWEEKLY",
        "descripcion": _("Biweekly Payroll - 15 days"),
        "dias": 15,
        "periodicidad": "quincenal",
        "periodos_por_anio": 24,
    },
    {
        "codigo": "FORTNIGHTLY",
        "descripcion": _("Fortnightly Payroll - 14 days"),
        "dias": 14,
        "periodicidad": "catorcenal",
        "periodos_por_anio": 26,
    },
    {
        "codigo": "WEEKLY",
        "descripcion": _("Weekly Payroll - 7 days"),
        "dias": 7,
        "periodicidad": "semanal",
        "periodos_por_anio": 52,
    },
]


def load_currencies() -> None:
    """Load American currencies into the database.

    Currency names are translated based on the configured language in the database.
    This function is idempotent - it will not create duplicates on repeated calls.
    """
    from coati_payroll.model import Moneda, db
    from coati_payroll.log import log

    currencies_loaded = 0
    for currency_data in CURRENCIES:
        # Check if currency already exists
        existing = db.session.execute(db.select(Moneda).filter_by(codigo=currency_data["codigo"])).scalar_one_or_none()

        if existing is None:
            # Create new currency - nombre will be translated by _()
            # Convert lazy string to regular string for database storage
            currency = Moneda()
            currency.codigo = currency_data["codigo"]
            currency.nombre = str(currency_data["nombre"])
            currency.simbolo = currency_data["simbolo"]
            currency.activo = True

            db.session.add(currency)
            currencies_loaded += 1

    if currencies_loaded > 0:
        db.session.commit()
        log.trace(f"Loaded {currencies_loaded} currencies")
    else:
        log.trace("No new currencies to load")


def load_income_concepts() -> None:
    """Load income concepts (percepciones) into the database.

    Concept names and descriptions are translated based on the configured
    language in the database. This function is idempotent.
    Default concepts are created with approved status.
    """
    from coati_payroll.model import Percepcion, db, utc_now
    from coati_payroll.enums import EstadoAprobacion
    from coati_payroll.log import log

    concepts_loaded = 0
    for concept_data in INCOME_CONCEPTS:
        # Check if concept already exists
        existing = db.session.execute(
            db.select(Percepcion).filter_by(codigo=concept_data["codigo"])
        ).scalar_one_or_none()

        if existing is None:
            # Create new income concept - strings will be translated by _()
            # Convert lazy strings to regular strings for database storage
            concept = Percepcion()
            concept.codigo = concept_data["codigo"]
            concept.nombre = str(concept_data["nombre"])
            concept.descripcion = str(concept_data["descripcion"])
            concept.formula_tipo = "fijo"
            concept.gravable = True
            concept.recurrente = False
            concept.activo = True
            concept.editable_en_nomina = True
            # Default concepts are pre-approved
            concept.estado_aprobacion = EstadoAprobacion.APROBADO
            concept.aprobado_por = "system"
            concept.aprobado_en = utc_now()
            concept.creado_por = "system"

            db.session.add(concept)
            concepts_loaded += 1

    if concepts_loaded > 0:
        db.session.commit()
        log.trace(f"Loaded {concepts_loaded} income concepts")
    else:
        log.trace("No new income concepts to load")


def load_deduction_concepts() -> None:
    """Load deduction concepts (deducciones) into the database.

    Concept names and descriptions are translated based on the configured
    language in the database. This function is idempotent.
    Default concepts are created with approved status.
    """
    from coati_payroll.model import Deduccion, db, utc_now
    from coati_payroll.enums import EstadoAprobacion
    from coati_payroll.log import log

    concepts_loaded = 0
    for concept_data in DEDUCTION_CONCEPTS:
        # Check if concept already exists
        existing = db.session.execute(
            db.select(Deduccion).filter_by(codigo=concept_data["codigo"])
        ).scalar_one_or_none()

        if existing is None:
            # Create new deduction concept - strings will be translated by _()
            # Convert lazy strings to regular strings for database storage
            concept = Deduccion()
            concept.codigo = concept_data["codigo"]
            concept.nombre = str(concept_data["nombre"])
            concept.descripcion = str(concept_data["descripcion"])
            concept.tipo = "general"
            concept.es_impuesto = False
            concept.formula_tipo = "fijo"
            concept.antes_impuesto = False
            concept.recurrente = False
            concept.activo = True
            concept.editable_en_nomina = True
            # Default concepts are pre-approved
            concept.estado_aprobacion = EstadoAprobacion.APROBADO
            concept.aprobado_por = "system"
            concept.aprobado_en = utc_now()
            concept.creado_por = "system"

            db.session.add(concept)
            concepts_loaded += 1

    if concepts_loaded > 0:
        db.session.commit()
        log.trace(f"Loaded {concepts_loaded} deduction concepts")
    else:
        log.trace("No new deduction concepts to load")


def load_benefit_concepts() -> None:
    """Load employer benefit concepts (prestaciones) into the database.

    Concept names and descriptions are translated based on the configured
    language in the database. This function is idempotent.
    Default concepts are created with approved status.
    """
    from coati_payroll.model import Prestacion, db, utc_now
    from coati_payroll.enums import EstadoAprobacion
    from coati_payroll.log import log

    concepts_loaded = 0
    for concept_data in BENEFIT_CONCEPTS:
        # Check if concept already exists
        existing = db.session.execute(
            db.select(Prestacion).filter_by(codigo=concept_data["codigo"])
        ).scalar_one_or_none()

        if existing is None:
            # Create new benefit concept - strings will be translated by _()
            # Convert lazy strings to regular strings for database storage
            concept = Prestacion()
            concept.codigo = concept_data["codigo"]
            concept.nombre = str(concept_data["nombre"])
            concept.descripcion = str(concept_data["descripcion"])
            concept.tipo = "patronal"
            concept.formula_tipo = "fijo"
            concept.recurrente = False
            concept.activo = True
            concept.editable_en_nomina = True
            # Default concepts are pre-approved
            concept.estado_aprobacion = EstadoAprobacion.APROBADO
            concept.aprobado_por = "system"
            concept.aprobado_en = utc_now()
            concept.creado_por = "system"

            db.session.add(concept)
            concepts_loaded += 1

    if concepts_loaded > 0:
        db.session.commit()
        log.trace(f"Loaded {concepts_loaded} benefit concepts")
    else:
        log.trace("No new benefit concepts to load")


def load_payroll_types() -> None:
    """Load common payroll types into the database.

    Payroll type descriptions are translated based on the configured
    language in the database. This function is idempotent.
    """
    from coati_payroll.model import TipoPlanilla, db
    from coati_payroll.log import log

    types_loaded = 0
    for type_data in PAYROLL_TYPES:
        # Check if payroll type already exists
        existing = db.session.execute(
            db.select(TipoPlanilla).filter_by(codigo=type_data["codigo"])
        ).scalar_one_or_none()

        if existing is None:
            # Create new payroll type - strings will be translated by _()
            # Convert lazy strings to regular strings for database storage
            payroll_type = TipoPlanilla()
            payroll_type.codigo = type_data["codigo"]
            payroll_type.descripcion = str(type_data["descripcion"])
            payroll_type.dias = type_data["dias"]
            payroll_type.periodicidad = type_data["periodicidad"]
            payroll_type.periodos_por_anio = type_data["periodos_por_anio"]
            payroll_type.mes_inicio_fiscal = 1  # January
            payroll_type.dia_inicio_fiscal = 1  # 1st day
            payroll_type.acumula_anual = True
            payroll_type.activo = True

            db.session.add(payroll_type)
            types_loaded += 1

    if types_loaded > 0:
        db.session.commit()
        log.trace(f"Loaded {types_loaded} payroll types")
    else:
        log.trace("No new payroll types to load")


def load_liquidation_concepts() -> None:
    """Load liquidation concepts into the database.

    This function is idempotent.
    """
    from coati_payroll.model import LiquidacionConcepto, db
    from coati_payroll.log import log

    concepts_loaded = 0
    for concept_data in LIQUIDATION_CONCEPTS:
        existing = db.session.execute(
            db.select(LiquidacionConcepto).filter_by(codigo=concept_data["codigo"])
        ).scalar_one_or_none()

        if existing is None:
            concept = LiquidacionConcepto()
            concept.codigo = concept_data["codigo"]
            concept.nombre = str(concept_data["nombre"])
            concept.descripcion = str(concept_data["descripcion"])
            concept.activo = True
            db.session.add(concept)
            concepts_loaded += 1

    if concepts_loaded > 0:
        db.session.commit()
        log.trace(f"Loaded {concepts_loaded} liquidation concepts")
    else:
        log.trace("No new liquidation concepts to load")


def load_initial_data() -> None:
    """Load all initial data into the database.

    This function loads currencies, income concepts, deduction concepts,
    employer benefit concepts, and payroll types. All strings are translated
    based on the configured language in the database.
    This function is idempotent - safe to call multiple times.
    """
    from coati_payroll.log import log

    log.trace("Loading initial data")

    load_currencies()
    load_income_concepts()
    load_deduction_concepts()
    load_benefit_concepts()
    load_payroll_types()
    load_liquidation_concepts()

    log.trace("Initial data loading completed")



# Copyright 2025 BMO Soluciones, S.A.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Custom employee field CRUD routes."""

from __future__ import annotations

from flask import Blueprint, flash, redirect, render_template, request, url_for
from flask_login import current_user

from coati_payroll.forms import CustomFieldForm
from coati_payroll.i18n import _
from coati_payroll.rbac import require_read_access, require_write_access
from coati_payroll.model import CampoPersonalizado, db
from coati_payroll.vistas.constants import PER_PAGE

custom_field_bp = Blueprint("custom_field", __name__, url_prefix="/custom_field")


@custom_field_bp.route("/")
@require_read_access()
def index():
    """List all custom fields with pagination."""
    page = request.args.get("page", 1, type=int)
    pagination = db.paginate(
        db.select(CampoPersonalizado).order_by(CampoPersonalizado.orden),
        page=page,
        per_page=PER_PAGE,
        error_out=False,
    )
    return render_template(
        "modules/custom_field/index.html",
        custom_fields=pagination.items,
        pagination=pagination,
    )


@custom_field_bp.route("/new", methods=["GET", "POST"])
@require_write_access()
def new():
    """Create a new custom field."""
    form = CustomFieldForm()

    if form.validate_on_submit():
        custom_field = CampoPersonalizado()
        custom_field.nombre_campo = form.nombre_campo.data
        custom_field.etiqueta = form.etiqueta.data
        custom_field.tipo_dato = form.tipo_dato.data
        custom_field.descripcion = form.descripcion.data
        custom_field.orden = int(form.orden.data or 0)
        custom_field.activo = form.activo.data
        custom_field.creado_por = current_user.usuario

        db.session.add(custom_field)
        db.session.commit()
        flash(_("Campo personalizado creado exitosamente."), "success")
        return redirect(url_for("custom_field.index"))

    return render_template(
        "modules/custom_field/form.html",
        form=form,
        title=_("Nuevo Campo Personalizado"),
    )


@custom_field_bp.route("/edit/<string:id>", methods=["GET", "POST"])
@require_write_access()
def edit(id: str):
    """Edit an existing custom field."""
    custom_field = db.session.get(CampoPersonalizado, id)
    if not custom_field:
        flash(_("Campo personalizado no encontrado."), "error")
        return redirect(url_for("custom_field.index"))

    form = CustomFieldForm(obj=custom_field)

    if form.validate_on_submit():
        custom_field.nombre_campo = form.nombre_campo.data
        custom_field.etiqueta = form.etiqueta.data
        custom_field.tipo_dato = form.tipo_dato.data
        custom_field.descripcion = form.descripcion.data
        custom_field.orden = int(form.orden.data or 0)
        custom_field.activo = form.activo.data
        custom_field.modificado_por = current_user.usuario

        db.session.commit()
        flash(_("Campo personalizado actualizado exitosamente."), "success")
        return redirect(url_for("custom_field.index"))

    return render_template(
        "modules/custom_field/form.html",
        form=form,
        title=_("Editar Campo Personalizado"),
        custom_field=custom_field,
    )


@custom_field_bp.route("/delete/<string:id>", methods=["POST"])
@require_write_access()
def delete(id: str):
    """Delete a custom field."""
    custom_field = db.session.get(CampoPersonalizado, id)
    if not custom_field:
        flash(_("Campo personalizado no encontrado."), "error")
        return redirect(url_for("custom_field.index"))

    db.session.delete(custom_field)
    db.session.commit()
    flash(_("Campo personalizado eliminado exitosamente."), "success")
    return redirect(url_for("custom_field.index"))



# Copyright 2025 BMO Soluciones, S.A.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Auth module."""

from __future__ import annotations

# <-------------------------------------------------------------------------> #
# Standard library
# <-------------------------------------------------------------------------> #
from datetime import datetime

# <-------------------------------------------------------------------------> #
# Third party libraries
# <-------------------------------------------------------------------------> #
from argon2 import PasswordHasher
from argon2.exceptions import VerifyMismatchError
from flask import Blueprint, render_template, redirect, url_for, flash
from flask_login import login_user, logout_user

# <-------------------------------------------------------------------------> #
# Local modules
# <-------------------------------------------------------------------------> #
from coati_payroll.log import log
from coati_payroll.model import Usuario, database
from coati_payroll.forms import LoginForm
from coati_payroll.i18n import _

auth = Blueprint("auth", __name__)


@auth.route("/login", methods=["GET", "POST"])
def login():
    """Mostrar y procesar el formulario de inicio de sesión.

    Rate limited to 5 attempts per minute per IP address to prevent
    brute force attacks on user credentials. Rate limiting is configured
    in coati_payroll/__init__.py using Flask-Limiter.
    """
    form = LoginForm()

    if form.validate_on_submit():
        usuario_id = form.email.data or ""
        clave = form.password.data or ""

        if validar_acceso(usuario_id, clave):
            # Cargar el registro del usuario (puede buscar por usuario o correo)
            registro = database.session.execute(
                database.select(Usuario).filter_by(usuario=usuario_id)
            ).scalar_one_or_none()

            if not registro:
                registro = database.session.execute(
                    database.select(Usuario).filter_by(correo_electronico=usuario_id)
                ).scalar_one_or_none()

            if registro is not None:
                login_user(registro)
                return redirect(url_for("app.index"))

        # Si llegamos aquí, el login falló
        flash(_("Usuario o contraseña incorrectos."), "error")

    return render_template("auth/login.html", form=form)


@auth.route("/logout")
def logout():
    """Cerrar sesión del usuario."""
    logout_user()
    flash(_("Sesión cerrada correctamente."), "info")
    return redirect(url_for("auth.login"))


# ---------------------------------------------------------------------------------------
# Proteger contraseñas de usuarios.
# ---------------------------------------------------------------------------------------
ph = PasswordHasher()


def proteger_passwd(clave: str, /) -> bytes:
    """Devuelve una contraseña salteada con argon2."""
    _hash = ph.hash(clave.encode()).encode("utf-8")

    return _hash


def validar_acceso(usuario_id: str, acceso: str, /) -> bool:
    """Verifica el inicio de sesión del usuario."""
    log.trace(f"Verifying access for {usuario_id}")
    registro = database.session.execute(database.select(Usuario).filter_by(usuario=usuario_id)).scalar_one_or_none()

    if not registro:
        registro = database.session.execute(
            database.select(Usuario).filter_by(correo_electronico=usuario_id)
        ).scalar_one_or_none()

    if registro is not None:
        try:
            ph.verify(registro.acceso, acceso.encode())
            clave_validada = True
        except VerifyMismatchError:
            clave_validada = False
    else:
        log.trace(f"User record not found for {usuario_id}")
        clave_validada = False

    log.trace(f"Access validation result is {clave_validada}")
    if clave_validada:
        registro.ultimo_acceso = datetime.now()
        database.session.commit()

    return clave_validada



# Copyright 2025 BMO Soluciones, S.A.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""User CRUD routes."""

from __future__ import annotations

from flask import Blueprint, flash, redirect, render_template, request, url_for
from flask_login import current_user, login_required

from coati_payroll.auth import proteger_passwd, validar_acceso
from coati_payroll.enums import TipoUsuario
from coati_payroll.forms import UserForm, ProfileForm
from coati_payroll.i18n import _
from coati_payroll.model import Usuario, db
from coati_payroll.rbac import require_role
from coati_payroll.vistas.constants import PER_PAGE

user_bp = Blueprint("user", __name__, url_prefix="/user")


@user_bp.route("/")
@require_role(TipoUsuario.ADMIN)
def index():
    """List all users with pagination. Only administrators can manage users."""
    page = request.args.get("page", 1, type=int)
    pagination = db.paginate(
        db.select(Usuario).order_by(Usuario.usuario),
        page=page,
        per_page=PER_PAGE,
        error_out=False,
    )
    return render_template("modules/user/index.html", users=pagination.items, pagination=pagination)


@user_bp.route("/new", methods=["GET", "POST"])
@require_role(TipoUsuario.ADMIN)
def new():
    """Create a new user. Only administrators can create users."""
    form = UserForm()

    if form.validate_on_submit():
        user = Usuario()
        user.usuario = form.usuario.data
        if form.password.data:
            user.acceso = proteger_passwd(form.password.data)
        else:
            flash(_("La contraseña es requerida para nuevos usuarios."), "error")
            return render_template("modules/user/form.html", form=form, title=_("Nuevo Usuario"))

        user.nombre = form.nombre.data
        user.apellido = form.apellido.data
        user.correo_electronico = form.correo_electronico.data
        user.tipo = form.tipo.data
        user.activo = form.activo.data
        user.creado_por = current_user.usuario

        db.session.add(user)
        db.session.commit()
        flash(_("Usuario creado exitosamente."), "success")
        return redirect(url_for("user.index"))

    return render_template("modules/user/form.html", form=form, title=_("Nuevo Usuario"))


@user_bp.route("/edit/<string:id>", methods=["GET", "POST"])
@require_role(TipoUsuario.ADMIN)
def edit(id: str):
    """Edit an existing user. Only administrators can edit users."""
    user = db.session.get(Usuario, id)
    if not user:
        flash(_("Usuario no encontrado."), "error")
        return redirect(url_for("user.index"))

    form = UserForm(obj=user)

    if form.validate_on_submit():
        user.usuario = form.usuario.data
        if form.password.data:
            user.acceso = proteger_passwd(form.password.data)
        user.nombre = form.nombre.data
        user.apellido = form.apellido.data
        user.correo_electronico = form.correo_electronico.data
        user.tipo = form.tipo.data
        user.activo = form.activo.data
        user.modificado_por = current_user.usuario

        db.session.commit()
        flash(_("Usuario actualizado exitosamente."), "success")
        return redirect(url_for("user.index"))

    # Don't show password in form
    form.password.data = ""
    return render_template("modules/user/form.html", form=form, title=_("Editar Usuario"), user=user)


@user_bp.route("/delete/<string:id>", methods=["POST"])
@require_role(TipoUsuario.ADMIN)
def delete(id: str):
    """Delete a user. Only administrators can delete users."""
    user = db.session.get(Usuario, id)
    if not user:
        flash(_("Usuario no encontrado."), "error")
        return redirect(url_for("user.index"))

    if user.id == current_user.id:
        flash(_("No puedes eliminar tu propio usuario."), "error")
        return redirect(url_for("user.index"))

    db.session.delete(user)
    db.session.commit()
    flash(_("Usuario eliminado exitosamente."), "success")
    return redirect(url_for("user.index"))


@user_bp.route("/profile", methods=["GET", "POST"])
@login_required
def profile():
    """Edit current user's profile and password."""
    form = ProfileForm(obj=current_user)

    if form.validate_on_submit():
        # Update basic profile information
        current_user.nombre = form.nombre.data
        current_user.apellido = form.apellido.data
        current_user.correo_electronico = form.correo_electronico.data

        # Handle password change if any password field has data
        password_change_attempted = form.current_password.data or form.new_password.data or form.confirm_password.data

        if password_change_attempted:
            # Validate password change
            error_message = None

            if not form.current_password.data:
                error_message = _("Debe ingresar la contraseña actual.")
            elif not validar_acceso(current_user.usuario, form.current_password.data):
                error_message = _("La contraseña actual es incorrecta.")
            elif not form.new_password.data:
                error_message = _("Debe ingresar una nueva contraseña.")
            elif form.new_password.data != form.confirm_password.data:
                error_message = _("Las contraseñas no coinciden.")

            if error_message:
                flash(error_message, "error")
                return render_template("modules/user/profile.html", form=form)

            # Update password
            current_user.acceso = proteger_passwd(form.new_password.data)
            flash(_("Contraseña actualizada exitosamente."), "success")

        current_user.modificado_por = current_user.usuario
        db.session.commit()
        flash(_("Perfil actualizado exitosamente."), "success")
        return redirect(url_for("user.profile"))

    # Clear password fields
    form.current_password.data = ""
    form.new_password.data = ""
    form.confirm_password.data = ""

    return render_template("modules/user/profile.html", form=form)



# Copyright 2025 BMO Soluciones, S.A.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Base configuration for the payroll module."""

from __future__ import annotations

# <-------------------------------------------------------------------------> #
# Standard library
# <-------------------------------------------------------------------------> #
from os import environ, getcwd, path, sys
from pathlib import Path
from urllib.parse import parse_qs, urlencode, urlparse, urlunparse

# <-------------------------------------------------------------------------> #
# Third party libraries
# <-------------------------------------------------------------------------> #


# <-------------------------------------------------------------------------> #
# Local modules
# <-------------------------------------------------------------------------> #
from coati_payroll.log import log


# ---------------------------------------------------------------------------------------
# Configuration file loading functionality
# ---------------------------------------------------------------------------------------
def load_config_from_file() -> dict:
    """
    Busca y carga configuración desde archivo con ConfigObj.

    Busca en las siguientes ubicaciones en orden:
    1. /etc/coati-payroll/coati-payroll.conf
    2. /etc/coati-payroll.conf
    3. ~/.config/coati-payroll/coati-payroll.conf
    3. ./coati-payroll.conf

    Returns:
        dict: Diccionario de configuración, vacío si no se encuentra archivo
    """
    try:
        from configobj import ConfigObj
    except ImportError:
        log.debug("ConfigObj not available, skipping file-based configuration.")
        return {}

    search_paths = [
        "/etc/coati-payroll/coati-payroll.conf",
        "/etc/coati-payroll.conf",
        path.expanduser("~/.config/coati-payroll/coati-payroll.conf"),
        path.join(getcwd(), "coati-payroll.conf"),
    ]

    for config_path in search_paths:
        if config_path and path.isfile(config_path):
            try:
                log.info(f"Loading configuration from file: {config_path}")
                config_obj = ConfigObj(config_path, encoding="utf-8")

                config_dict = dict(config_obj)

                if "DATABASE_URL" in config_dict:
                    config_dict["SQLALCHEMY_DATABASE_URI"] = config_dict["DATABASE_URL"]

                if "REDIS_URL" in config_dict:
                    config_dict["CACHE_REDIS_URL"] = config_dict["REDIS_URL"]

                return config_dict

            except Exception as e:
                log.warning(f"Error loading configuration from {config_path}: {e}")
                continue

    log.trace("No configuration file found in search paths.")
    return {}


# < --------------------------------------------------------------------------------------------- >
# Configuración central de la aplicación.
VALORES_TRUE = {*["1", "true", "yes", "on"], *["development", "dev"]}
DEBUG_VARS = ["DEBUG", "CI", "DEV", "DEVELOPMENT"]
FRAMEWORK_VARS = ["FLASK_ENV", "DJANGO_DEBUG", "NODE_ENV"]
GENERIC_VARS = ["ENV", "APP_ENV"]

# < --------------------------------------------------------------------------------------------- >
# Gestión de variables de entorno.
DESARROLLO = any(
    str(environ.get(var, "")).strip().lower() in VALORES_TRUE for var in [*DEBUG_VARS, *FRAMEWORK_VARS, *GENERIC_VARS]
)

# < --------------------------------------------------------------------------------------------- >
# Directorios base de la aplicacion
DIRECTORIO_ACTUAL: Path = Path(path.abspath(path.dirname(__file__)))
DIRECTORIO_APP: Path = DIRECTORIO_ACTUAL.parent.absolute()
DIRECTORIO_DESARROLLO: Path = DIRECTORIO_APP
DIRECTORIO_PLANTILLAS_BASE: str = path.join(DIRECTORIO_ACTUAL, "templates")
DIRECTORIO_ARCHIVOS_BASE: str = path.join(DIRECTORIO_ACTUAL, "static")

# < --------------------------------------------------------------------------------------------- >
# Directorios personalizados para la aplicación.
DIRECTORIO_ARCHIVOS = DIRECTORIO_ARCHIVOS_BASE
DIRECTORIO_PLANTILLAS = DIRECTORIO_PLANTILLAS_BASE


# < --------------------------------------------------------------------------------------------- >
TESTING = (
    "PYTEST_CURRENT_TEST" in environ
    or "PYTEST_VERSION" in environ
    or "TESTING" in environ
    or hasattr(sys, "_called_from_test")
    or environ.get("CI")
    or "pytest" in sys.modules
    or path.basename(sys.argv[0]) in ["pytest", "py.test"]
)

if TESTING:
    # Use DB in memory for tests to avoid filesystem side-effects
    SQLITE: str = "sqlite:///:memory:"
else:
    # File-based sqlite in project root
    sqlite_file = DIRECTORIO_DESARROLLO.joinpath("coati_payroll.db")
    SQLITE = f"sqlite:///{sqlite_file.as_posix()}"

# < --------------------------------------------------------------------------------------------- >
# Configuración de la aplicación:
# Se siguen las recomendaciones de "Twelve Factors App" y las opciones se leen del entorno.
CONFIGURACION: dict[str, str | bool | Path] = {}
CONFIGURACION["SECRET_KEY"] = environ.get("SECRET_KEY") or "dev"  # nosec
CONFIGURACION["SQLALCHEMY_DATABASE_URI"] = environ.get("DATABASE_URL") or SQLITE  # nosec
# Opciones comunes de configuración.
CONFIGURACION["PRESERVE_CONTEXT_ON_EXCEPTION"] = False

if DESARROLLO:
    log.warning("Using default configuration.")
    log.info("Default configuration is not recommended for use in production environments.")
    CONFIGURACION["SQLALCHEMY_TRACK_MODIFICATIONS"] = "False"
    CONFIGURACION["TEMPLATES_AUTO_RELOAD"] = True

# < --------------------------------------------------------------------------------------------- >
# Corrige la URI de conexión a la base de datos si el usuario omite el driver apropiado.

if DATABASE_URL_BASE := CONFIGURACION.get("SQLALCHEMY_DATABASE_URI"):

    DATABASE_URL_CORREGIDA = DATABASE_URL_BASE

    prefix = DATABASE_URL_BASE.split(":", 1)[0]  # Extraer prefijo: "postgres", "mysql", etc.

    # Caso especial: Heroku + PostgreSQL
    if environ.get("DYNO") and prefix in (
        "postgres",
        "postgresql",
    ):  # type: ignore[operator]
        parsed = urlparse(DATABASE_URL_BASE)  # type: ignore[arg-type]
        query = parse_qs(parsed.query)
        query["sslmode"] = ["require"]
        DATABASE_URL_CORREGIDA = urlunparse(parsed._replace(scheme="postgresql", query=urlencode(query, doseq=True)))

    else:
        # Corrige el esquema según el prefijo detectado
        match prefix:
            case "postgresql":
                parsed = urlparse(DATABASE_URL_BASE)
                query = parse_qs(parsed.query)
                query.pop("sslmode", None)
                new_query = urlencode(query, doseq=True) if query else ""
                cleaned = urlunparse(parsed._replace(query=new_query))
                DATABASE_URL_CORREGIDA = "postgresql+pg8000" + cleaned[10:]
            case "postgres":
                parsed = urlparse(DATABASE_URL_BASE)
                query = parse_qs(parsed.query)
                query.pop("sslmode", None)
                new_query = urlencode(query, doseq=True) if query else ""
                cleaned = urlunparse(parsed._replace(query=new_query))
                DATABASE_URL_CORREGIDA = "postgresql+pg8000" + cleaned[8:]
            case "mysql":
                DATABASE_URL_CORREGIDA = "mysql+pymysql" + DATABASE_URL_BASE[5:]
            case "mariadb":
                DATABASE_URL_CORREGIDA = "mariadb+mariadbconnector" + DATABASE_URL_BASE[7:]
            case _:
                pass

    # Actualizar configuración si hubo cambio
    if DATABASE_URL_BASE != DATABASE_URL_CORREGIDA:
        log.info(f"Database URI corrected: {DATABASE_URL_BASE} → {DATABASE_URL_CORREGIDA}")
        CONFIGURACION["SQLALCHEMY_DATABASE_URI"] = DATABASE_URL_CORREGIDA

# < --------------------------------------------------------------------------------------------- >
# Queue configuration for background job processing
# The system will automatically select between Dramatiq (Redis) and Huey (filesystem)
# based on REDIS_URL availability
CONFIGURACION["QUEUE_ENABLED"] = environ.get("QUEUE_ENABLED", "1") in [
    "1",
    "true",
    "True",
    "yes",
]
CONFIGURACION["QUEUE_STORAGE_PATH"] = environ.get("COATI_QUEUE_PATH")  # For Huey filesystem

# Background payroll processing configuration
# Threshold for automatic background processing (number of employees)
# Payrolls with more employees than this threshold will be processed in background
# Default: 100 employees. Can be adjusted based on system performance:
# - For systems with complex formulas or slow performance: lower to 50 or 25
# - For high-performance systems: increase to 200 or 500
CONFIGURACION["BACKGROUND_PAYROLL_THRESHOLD"] = int(environ.get("BACKGROUND_PAYROLL_THRESHOLD", "100"))

# < --------------------------------------------------------------------------------------------- >
configuration = CONFIGURACION



# Copyright 2025 BMO Soluciones, S.A.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Service for novedad business logic."""

from decimal import Decimal
from coati_payroll.model import db, Planilla, Nomina, NominaNovedad
from coati_payroll.vistas.planilla.helpers.form_helpers import get_concepto_ids_from_form


class NovedadService:
    """Service for novedad operations."""

    @staticmethod
    def listar_novedades(planilla: Planilla, nomina: Nomina) -> list:
        """List all novedades for a nomina.

        Args:
            planilla: The planilla
            nomina: The nomina

        Returns:
            List of NominaNovedad objects
        """
        # Get all employees in this planilla
        empleado_ids = [pe.empleado_id for pe in planilla.planilla_empleados if pe.activo]

        # Query novedades that fall within the nomina period and are for employees in this planilla
        novedades = (
            db.session.execute(
                db.select(NominaNovedad)
                .filter(
                    NominaNovedad.empleado_id.in_(empleado_ids),
                    NominaNovedad.fecha_novedad >= nomina.periodo_inicio,
                    NominaNovedad.fecha_novedad <= nomina.periodo_fin,
                )
                .order_by(NominaNovedad.fecha_novedad.desc(), NominaNovedad.timestamp.desc())
            )
            .scalars()
            .all()
        )

        return novedades

    @staticmethod
    def validar_fecha_novedad(fecha_novedad, nomina: Nomina) -> tuple[bool, str | None]:
        """Validate that fecha_novedad falls within the nomina period.

        Args:
            fecha_novedad: The date to validate
            nomina: The nomina

        Returns:
            Tuple of (is_valid, error_message)
        """
        from coati_payroll.i18n import _

        if fecha_novedad:
            if fecha_novedad < nomina.periodo_inicio or fecha_novedad > nomina.periodo_fin:
                return False, _(
                    "La fecha de la novedad debe estar dentro del período de la nómina " "({} a {})."
                ).format(
                    nomina.periodo_inicio.strftime("%d/%m/%Y"),
                    nomina.periodo_fin.strftime("%d/%m/%Y"),
                )
        return True, None

    @staticmethod
    def crear_novedad(nomina: Nomina, form, usuario: str) -> NominaNovedad:
        """Create a new novedad.

        Args:
            nomina: The nomina
            form: The form with novedad data
            usuario: Username of the user creating

        Returns:
            The created NominaNovedad
        """
        percepcion_id, deduccion_id = get_concepto_ids_from_form(form)

        novedad = NominaNovedad(
            nomina_id=nomina.id,
            empleado_id=form.empleado_id.data,
            codigo_concepto=form.codigo_concepto.data,
            tipo_valor=form.tipo_valor.data,
            valor_cantidad=Decimal(str(form.valor_cantidad.data)),
            fecha_novedad=form.fecha_novedad.data,
            percepcion_id=percepcion_id,
            deduccion_id=deduccion_id,
            creado_por=usuario,
        )
        db.session.add(novedad)
        db.session.commit()
        return novedad

    @staticmethod
    def actualizar_novedad(novedad: NominaNovedad, form, usuario: str):
        """Update an existing novedad.

        Args:
            novedad: The novedad to update
            form: The form with updated data
            usuario: Username of the user updating
        """
        percepcion_id, deduccion_id = get_concepto_ids_from_form(form)

        novedad.empleado_id = form.empleado_id.data
        novedad.codigo_concepto = form.codigo_concepto.data
        novedad.tipo_valor = form.tipo_valor.data
        novedad.valor_cantidad = Decimal(str(form.valor_cantidad.data))
        novedad.fecha_novedad = form.fecha_novedad.data
        novedad.percepcion_id = percepcion_id
        novedad.deduccion_id = deduccion_id
        novedad.modificado_por = usuario
        db.session.commit()



# Copyright 2025 BMO Soluciones, S.A.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Repository for AcumuladoAnual operations."""

from __future__ import annotations

from datetime import date
from typing import Optional

from coati_payroll.model import AcumuladoAnual, Empleado
from .base_repository import BaseRepository


class AcumuladoRepository(BaseRepository[AcumuladoAnual]):
    """Repository for AcumuladoAnual operations."""

    def get_by_id(self, acumulado_id: str) -> Optional[AcumuladoAnual]:
        """Get acumulado by ID."""
        return self.session.get(AcumuladoAnual, acumulado_id)

    def get_or_create(
        self,
        empleado: Empleado,
        tipo_planilla_id: str,
        empresa_id: str,
        periodo_fiscal_inicio: date,
    ) -> AcumuladoAnual:
        """Get or create acumulado for employee and fiscal period."""
        from sqlalchemy import select

        acumulado = (
            self.session.execute(
                select(AcumuladoAnual).filter(
                    AcumuladoAnual.empleado_id == empleado.id,
                    AcumuladoAnual.tipo_planilla_id == tipo_planilla_id,
                    AcumuladoAnual.empresa_id == empresa_id,
                    AcumuladoAnual.periodo_fiscal_inicio == periodo_fiscal_inicio,
                )
            )
            .unique()
            .scalar_one_or_none()
        )

        if not acumulado:
            from datetime import date as date_type
            from decimal import Decimal

            periodo_fiscal_fin = date_type(
                periodo_fiscal_inicio.year + 1, periodo_fiscal_inicio.month, periodo_fiscal_inicio.day
            )

            acumulado = AcumuladoAnual(
                empleado_id=empleado.id,
                tipo_planilla_id=tipo_planilla_id,
                empresa_id=empresa_id,
                periodo_fiscal_inicio=periodo_fiscal_inicio,
                periodo_fiscal_fin=periodo_fiscal_fin,
                salario_bruto_acumulado=empleado.salario_acumulado or Decimal("0.00"),
                salario_gravable_acumulado=Decimal("0.00"),
                deducciones_antes_impuesto_acumulado=empleado.impuesto_acumulado or Decimal("0.00"),
                impuesto_retenido_acumulado=Decimal("0.00"),
                periodos_procesados=0,
                salario_acumulado_mes=Decimal("0.00"),
            )
            self.session.add(acumulado)

        return acumulado

    def save(self, acumulado: AcumuladoAnual) -> AcumuladoAnual:
        """Save acumulado."""
        self.session.add(acumulado)
        return acumulado



# Copyright 2025 BMO Soluciones, S.A.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""App module."""

from __future__ import annotations

# <-------------------------------------------------------------------------> #
# Standard library
# <-------------------------------------------------------------------------> #

# <-------------------------------------------------------------------------> #
# Third party libraries
# <-------------------------------------------------------------------------> #
from flask import Blueprint, render_template
from flask_login import login_required
from sqlalchemy import func
from sqlalchemy.exc import SQLAlchemyError

# <-------------------------------------------------------------------------> #
# Local modules
# <-------------------------------------------------------------------------> #
from coati_payroll.model import db, Empleado, Empresa, Planilla, Nomina

app = Blueprint("app", __name__)


@app.route("/")
@login_required
def index():
    # Get statistics for dashboard
    total_empleados = (
        db.session.execute(db.select(func.count(Empleado.id)).filter(Empleado.activo.is_(True))).scalar() or 0
    )
    total_empresas = (
        db.session.execute(db.select(func.count(Empresa.id)).filter(Empresa.activo.is_(True))).scalar() or 0
    )
    total_planillas = (
        db.session.execute(db.select(func.count(Planilla.id)).filter(Planilla.activo.is_(True))).scalar() or 0
    )
    total_nominas = db.session.execute(db.select(func.count(Nomina.id))).scalar() or 0

    # Get recent payrolls (last 5)
    recent_nominas = (
        db.session.execute(db.select(Nomina).order_by(Nomina.fecha_generacion.desc()).limit(5)).scalars().all()
    )

    return render_template(
        "index.html",
        total_empleados=total_empleados,
        total_empresas=total_empresas,
        total_planillas=total_planillas,
        total_nominas=total_nominas,
        recent_nominas=recent_nominas,
    )


@app.route("/health")
def health():
    """Health check endpoint for container orchestration.

    Returns a simple OK response to indicate the application is running.
    This endpoint does not require authentication and does not check database connectivity.
    """
    return {"status": "ok"}, 200


@app.route("/ready")
def ready():
    """Readiness check endpoint for container orchestration.

    Returns OK if the application is ready to serve traffic (database is accessible).
    This endpoint does not require authentication.
    """
    try:
        # Test database connectivity using a fresh connection
        # This avoids issues with session state from other parts of the application
        with db.engine.connect() as connection:
            connection.execute(db.text("SELECT 1")).scalar()
        return {"status": "ok"}, 200
    except SQLAlchemyError:
        return {"status": "unavailable"}, 503
    except Exception:
        # Catch any other exception that might occur
        return {"status": "unavailable"}, 503
